import json
import logging
import http.client as http
from collections import namedtuple

from .common import *

logger = logging.getLogger(__name__)


# additional parameters are: tempLink='true'
def get_node_list(**params) -> list:
    return BackOffRequest.paginated_get(get_metadata_url() + 'nodes', params)


def get_file_list() -> list:
    return get_node_list(filters='kind:FILE')


def get_folder_list() -> list:
    return get_node_list(filters='kind:FOLDER')


def get_asset_list() -> list:
    return get_node_list(filters='kind:ASSET')


def get_trashed_folders() -> list:
    return get_node_list(filters='status:TRASH AND kind:FOLDER')


def get_trashed_files() -> list:
    return get_node_list(filters='status:TRASH AND kind:FILE')

ChangeSet = namedtuple('Changes', ['nodes', 'purged_nodes', 'checkpoint', 'reset'])


def get_changes(checkpoint='', include_purged=False) -> ChangeSet:
    """ https://developer.amazon.com/public/apis/experience/cloud-drive/content/changes
    :returns ChangeSet: list of nodes, list of purged nodes, last checkpoint, reset flag
    """

    logger.info('Getting changes with checkpoint "%s".' % checkpoint)

    body = {}
    if checkpoint:
        body['checkpoint'] = checkpoint
    if include_purged:
        body['includePurged'] = 'true'
    r = BackOffRequest.post(get_metadata_url() + 'changes', data=json.dumps(body), stream=True)
    if r.status_code not in OK_CODES:
        r.close()
        raise RequestError(r.status_code, r.text)

    """ return format should be:
    {"checkpoint": str, "reset": bool, "nodes": []}
    {"checkpoint": str, "reset": false, "nodes": []}
    {"end": true}
    """
    reset = False
    nodes = []
    purged_nodes = []

    end = False
    pages = -1

    for line in r.iter_lines(chunk_size=10 * 1024 ** 2, decode_unicode=False):
        # filter out keep-alive new lines
        if not line:
            continue

        pages += 1

        try:
            o = json.loads(line.decode('utf-8'))
        except ValueError:
            raise RequestError(RequestError.CODE.INCOMPLETE_RESULT, '[acd_cli] Invalid JSON in change set.')

        try:
            if o['end']:
                end = True
                continue
        except KeyError:
            pass

        if o['reset']:
            logger.info('Found "reset" tag in changes.')
            reset = True

        # could this actually happen?
        if o['statusCode'] not in OK_CODES:
            raise RequestError(RequestError.CODE.FAILED_SUBREQUEST, '[acd_cli] Partial failure in change request.')

        for node in o['nodes']:
            if node['status'] == 'PURGED':
                purged_nodes.append(node['id'])
            else:
                nodes.append(node)
        checkpoint = o['checkpoint']

    r.close()

    logger.info('%i pages, %i nodes, %i purged nodes in changes.' % (pages, len(nodes), len(purged_nodes)))
    if not end:
        logger.warning('End of change request not reached.')

    return ChangeSet(nodes, purged_nodes, checkpoint, reset)


def get_metadata(node_id: str) -> dict:
    params = {'tempLink': 'true'}
    r = BackOffRequest.get(get_metadata_url() + 'nodes/' + node_id, params=params)
    if r.status_code not in OK_CODES:
        raise RequestError(r.status_code, r.text)
    return r.json()


# this will increment the node's version attribute
def update_metadata(node_id: str, properties: dict) -> dict:
    body = json.dumps(properties)
    r = BackOffRequest.patch(get_metadata_url() + 'nodes/' + node_id, data=body)
    if r.status_code not in OK_CODES:
        raise RequestError(r.status_code, r.text)
    return r.json()


def get_root_id() -> dict:
    params = {'filters': 'isRoot:true'}
    r = BackOffRequest.get(get_metadata_url() + 'nodes', params=params)

    if r.status_code not in OK_CODES:
        raise RequestError(r.status_code, r.text)

    data = r.json()

    if 'id' in data['data'][0]:
        return data['data'][0]['id']


def list_children(node_id: str) -> list:
    l = BackOffRequest.paginated_get(get_metadata_url() + 'nodes/' + node_id + '/children')
    return l


def add_child(parent_id: str, child_id: str) -> dict:
    """:returns updated child node dict"""
    r = BackOffRequest.put(get_metadata_url() + 'nodes/' + parent_id + '/children/' + child_id)
    if r.status_code not in OK_CODES:
        logger.error('Adding child failed.')
        raise RequestError(r.status_code, r.text)
    return r.json()


def remove_child(parent_id: str, child_id: str) -> dict:
    """:returns updated child node dict"""
    r = BackOffRequest.delete(get_metadata_url() + 'nodes/' + parent_id + "/children/" + child_id)
    # contrary to response code stated in API doc (202 ACCEPTED)
    if r.status_code not in OK_CODES:
        logger.error('Removing child failed.')
        raise RequestError(r.status_code, r.text)
    return r.json()


def move_node_from(child_id: str, old_parent_id, new_parent_id: str) -> dict:
    r = add_child(new_parent_id, child_id)
    return remove_child(old_parent_id, child_id)


def rename_node(node_id: str, new_name: str) -> dict:
    properties = {'name': new_name}
    return update_metadata(node_id, properties)


def set_available(node_id: str) -> dict:
    """Sets node status from 'PENDING' to 'AVAILABLE'."""
    properties = {'status': 'AVAILABLE'}
    return update_metadata(node_id, properties)


def list_properties(node_id: str, owner_id: str) -> dict:
    """This will always return an empty dict if the accessor is not the owner.
    :param owner_id: owner ID (return status 404 if empty)
    """
    r = BackOffRequest.get(get_metadata_url() + 'nodes/' + node_id + '/properties/' + owner_id)
    if r.status_code not in OK_CODES:
        raise RequestError(r.status_code, r.text)
    return r.json()['data']


def add_property(node_id: str, owner_id: str, key: str, value: str) -> dict:
    """Adds or overwrites property. Maximum number of keys per owner is 10.
    :param value: string of length <= 500
    """
    ok_codes = [http.CREATED]
    r = BackOffRequest.put(get_metadata_url() + 'nodes/' + node_id
                           + '/properties/' + owner_id + '/' + key,
                           data=json.dumps({'value': value}), acc_codes=ok_codes)
    if r.status_code not in ok_codes:
        raise RequestError(r.status_code, r.text)
    return r.json()


def delete_property(node_id: str, owner_id: str, key: str):
    ok_codes = [http.NO_CONTENT]
    r = BackOffRequest.delete(get_metadata_url() + 'nodes/' + node_id
                              + '/properties/' + owner_id + '/' + key, acc_codes=ok_codes)
    if r.status_code not in ok_codes:
        raise RequestError(r.status_code, r.text)


"""File and folder creation, file transfer operations."""

import http.client as http
import os
import json
import io
import mimetypes
from collections import OrderedDict
import logging
from urllib.parse import quote_plus

try:
    from requests_toolbelt import MultipartEncoder
except ImportError:
    from acdcli.bundled.encoder import MultipartEncoder

from .common import *

FS_RW_CHUNK_SZ = 1024 * 128

PARTIAL_SUFFIX = '.__incomplete'
CHUNK_SIZE = 500 * 1024 ** 2  # basically arbitrary
CHUNK_MAX_RETRY = 5
CONSECUTIVE_DL_LIMIT = CHUNK_SIZE

logger = logging.getLogger(__name__)


def tee_open(path: str, **kwargs):
    f = open(path, 'rb')
    return TeeBufferedReader(f, **kwargs)


class TeeBufferedReader(object):
    """Creates proxy buffered reader object that allows callbacks on read operations."""

    def __init__(self, file: io.BufferedReader, callbacks: list=None):
        self._file = file
        self._callbacks = callbacks

    def __getattr__(self, item):
        try:
            return object.__getattr__(item)
        except AttributeError:
            return getattr(self._file, item)

    def read(self, ln=-1):
        ln = ln if ln in (0, -1) else FS_RW_CHUNK_SZ
        chunk = self._file.read(ln)
        if self._callbacks:
            for callback in self._callbacks:
                callback(chunk)
        return chunk


def create_folder(name: str, parent=None) -> dict:
    body = {'kind': 'FOLDER', 'name': name}
    if parent:
        body['parents'] = [parent]
    body_str = json.dumps(body)

    acc_codes = [http.CREATED]

    r = BackOffRequest.post(get_metadata_url() + 'nodes', acc_codes=acc_codes, data=body_str)

    if r.status_code not in acc_codes:
        raise RequestError(r.status_code, r.text)

    return r.json()


def _get_mimetype(file_name: str) -> str:
    mt = mimetypes.guess_type(file_name)[0]
    return mt if mt else 'application/octet-stream'


def create_file(file_name: str, parent: str=None) -> dict:
    params = {'suppress': 'deduplication'}

    basename = os.path.basename(file_name)
    metadata = {'kind': 'FILE', 'name': basename}
    if parent:
        metadata['parents'] = [parent]
    mime_type = _get_mimetype(basename)
    f = io.BytesIO()

    # basename is ignored
    m = MultipartEncoder(fields=OrderedDict([('metadata', json.dumps(metadata)),
                                             ('content', (quote_plus(basename), f, mime_type))]))

    ok_codes = [http.CREATED]
    r = BackOffRequest.post(get_content_url() + 'nodes', params=params, data=m,
                            acc_codes=ok_codes, headers={'Content-Type': m.content_type})

    if r.status_code not in ok_codes:
        raise RequestError(r.status_code, r.text)
    return r.json()


def upload_file(file_name: str, parent: str=None, read_callbacks=None, deduplication=False) -> dict:
    params = {'suppress': 'deduplication'}
    if deduplication and os.path.getsize(file_name) > 0:
        params = {'suppress': 'deduplication'}

    basename = os.path.basename(file_name)
    metadata = {'kind': 'FILE', 'name': basename}
    if parent:
        metadata['parents'] = [parent]
    mime_type = _get_mimetype(basename)
    f = tee_open(file_name, callbacks=read_callbacks)

    # basename is ignored
    m = MultipartEncoder(fields=OrderedDict([('metadata', json.dumps(metadata)),
                                             ('content', (quote_plus(basename), f, mime_type))]))

    ok_codes = [http.CREATED]
    r = BackOffRequest.post(get_content_url() + 'nodes', params=params, data=m, acc_codes=ok_codes,
                            stream=True, headers={'Content-Type': m.content_type})

    if r.status_code not in ok_codes:
        raise RequestError(r.status_code, r.text)
    return r.json()


def overwrite_file(node_id: str, file_name: str, read_callbacks=None, deduplication=False) -> dict:
    params = {} if deduplication else {'suppress': 'deduplication'}

    basename = os.path.basename(file_name)
    mime_type = _get_mimetype(basename)
    f = tee_open(file_name, callbacks=read_callbacks)

    # basename is ignored
    m = MultipartEncoder(fields={('content', (quote_plus(basename), f, mime_type))})

    r = BackOffRequest.put(get_content_url() + 'nodes/' + node_id + '/content', params=params,
                           data=m,
                           stream=True, headers={'Content-Type': m.content_type})

    if r.status_code not in OK_CODES:
        raise RequestError(r.status_code, r.text)

    return r.json()


def download_file(node_id: str, basename: str, dirname: str=None, **kwargs):
    """ Deals with download preparation, download with :func:`chunked_download` and finish.
    Calls callbacks while fast forwarding through incomplete file (if existent).
    Will not check for existing file prior to download and overwrite existing file on finish.
    :param dirname: a valid local directory name, or CWD if None
    :param basename: a valid file name
    kwargs:
    length: the total length of the file
    write_callbacks (list[function]): passed on to :func:`chunked_download`
    resume (bool=True): whether to resume if partial file exists
    """

    dl_path = basename
    if dirname:
        dl_path = os.path.join(dirname, basename)
    part_path = dl_path + PARTIAL_SUFFIX
    offset = 0

    length = kwargs.get('length', 0)
    resume = kwargs.get('resume', True)
    if resume and os.path.isfile(part_path):
        with open(part_path, 'ab') as f:
            trunc_pos = os.path.getsize(part_path) - 1 - FS_RW_CHUNK_SZ
            f.truncate(trunc_pos if trunc_pos >= 0 else 0)

        write_callbacks = kwargs.get('write_callbacks')
        if write_callbacks:
            with open(part_path, 'rb') as f:
                for chunk in iter(lambda: f.read(FS_RW_CHUNK_SZ), b''):
                    for rcb in write_callbacks:
                        rcb(chunk)

        f = open(part_path, 'ab')
    else:
        f = open(part_path, 'wb')
    offset = f.tell()

    chunked_download(node_id, f, offset=offset, **kwargs)
    pos = f.tell()
    f.close()
    if length > 0:
        if pos < length:
            raise RequestError(RequestError.CODE.INCOMPLETE_RESULT, '[acd_cli] ')

    if os.path.isfile(dl_path):
        logger.info('Deleting existing file "%s".' % dl_path)
        os.remove(dl_path)
    os.rename(part_path, dl_path)


def response_chunk(node_id: str, offset: int, length: int, **kwargs):
    ok_codes = [http.PARTIAL_CONTENT]
    end = offset + length - 1
    logger.debug('o %d l %d' % (offset, length))

    r = BackOffRequest.get(get_content_url() + 'nodes/' + node_id + '/content',
                           acc_codes=ok_codes, stream=True,
                           headers={'Range': 'bytes=%d-%d' % (offset, end)}, **kwargs)
    # if r.status_code == http.REQUESTED_RANGE_NOT_SATISFIABLE:
    #     return
    if r.status_code not in ok_codes:
        raise RequestError(r.status_code, r.text)

    return r


def download_chunk(node_id: str, offset: int, length: int, **kwargs):
    """:param length: the length of the download chunk"""
    r = response_chunk(node_id, offset, length, **kwargs)
    if not r:
        return

    buffer = bytearray()
    for chunk in r.iter_content(chunk_size=FS_RW_CHUNK_SZ):
        if chunk:
            buffer.extend(chunk)
    return buffer


@catch_conn_exception
def chunked_download(node_id: str, file: io.BufferedWriter, **kwargs):
    """Keyword args:
    offset (int): byte offset -- start byte for ranged request
    length (int): total file length[!], equal to end + 1
    write_callbacks (list[function])
    """
    ok_codes = [http.PARTIAL_CONTENT]

    write_callbacks = kwargs.get('write_callbacks', [])

    chunk_start = kwargs.get('offset', 0)
    length = kwargs.get('length', 100 * 1024 ** 4)

    retries = 0
    while chunk_start < length:
        chunk_end = chunk_start + CHUNK_SIZE - 1
        if chunk_end >= length:
            chunk_end = length - 1

        if retries >= CHUNK_MAX_RETRY:
            raise RequestError(RequestError.CODE.FAILED_SUBREQUEST,
                               '[acd_cli] Downloading chunk failed multiple times.')
        r = BackOffRequest.get(get_content_url() + 'nodes/' + node_id + '/content', stream=True,
                               acc_codes=ok_codes,
                               headers={'Range': 'bytes=%d-%d' % (chunk_start, chunk_end)})

        logger.debug('Range %d-%d' % (chunk_start, chunk_end))
        # this should only happen at the end of unknown-length downloads
        if r.status_code == http.REQUESTED_RANGE_NOT_SATISFIABLE:
            logger.debug('Invalid byte range requested %d-%d' % (chunk_start, chunk_end))
            break
        if r.status_code not in ok_codes:
            r.close()
            retries += 1
            logging.debug('Chunk [%d-%d], retry %d.' % (chunk_start, chunk_end, retries))
            continue

        curr_ln = 0
        # connection exceptions occur here
        for chunk in r.iter_content(chunk_size=FS_RW_CHUNK_SZ):
            if chunk:  # filter out keep-alive new chunks
                file.write(chunk)
                file.flush()
                for wcb in write_callbacks:
                    wcb(chunk)
                curr_ln += len(chunk)
        chunk_start += CHUNK_SIZE
        retries = 0
        r.close()

    return



"""
This is a template that you can use for adding custom plugins.
May be subject to changes in the near future.
"""

from . import *


class TestPlugin(Plugin):
    MIN_VERSION = '0.1.3'

    @classmethod
    def attach(cls, subparsers: argparse.ArgumentParser, log: list, **kwargs):
        """ Attaches this plugin to the argparse action subparser group
        :param subparsers the action subparser group
        :param log a list to put initialization log messages in
         """
        p = subparsers.add_parser('test', add_help=False)
        p.add_argument('--silent', action='store_true', default=False)
        p.set_defaults(func=cls.action)

        log.append(str(cls) + ' attached.')

    @classmethod
    def action(cls, args: argparse.Namespace) -> int:
        """ This is where the magic happens. Return a zero for success, a non-zero int for failure. """
        if not args.silent:
            print('This plugin works.')
        return 0


import os
import time
import json
import http.client as http
import random
from time import sleep
import logging
import requests
from threading import Lock, local

from requests.exceptions import ConnectionError

try:
    from requests.exceptions import ReadTimeout as ReadTimeoutError
except ImportError:
    try:
        from requests.packages.urllib3.exceptions import ReadTimeoutError
    except ImportError:
        class ReadTimeoutError(Exception):
            pass

from . import oauth

__all__ = ('RequestError', 'BackOffRequest',
           'OK_CODES', 'init', 'catch_conn_exception', 'get_metadata_url', 'get_content_url')

logger = logging.getLogger(__name__)

# status codes that indicate request success
OK_CODES = [http.OK]

cache_path = ''
ENDPOINT_DATA_FILE = 'endpoint_data'
endpoint_data_path = lambda: os.path.join(cache_path, ENDPOINT_DATA_FILE)

# json key names
ACC_TOKEN_KEY = 'access_token'
REFR_TOKEN_KEY = 'refresh_token'

EXP_TIME_KEY = 'exp_time'

endpoint_data = {}
get_metadata_url = lambda: endpoint_data['metadataUrl']
get_content_url = lambda: endpoint_data['contentUrl']

AMZ_ENDPOINT_REQ_URL = 'https://drive.amazonaws.com/drive/v1/account/endpoint'
ENDPOINT_VAL_TIME = 259200

CONN_TIMEOUT = 30
IDLE_TIMEOUT = 60
REQUESTS_TIMEOUT = (CONN_TIMEOUT, IDLE_TIMEOUT) if requests.__version__ >= '2.4.0' else IDLE_TIMEOUT


def init(path='') -> bool:
    logger.info('Initializing acd with path "%s".' % path)

    global cache_path
    cache_path = path

    return oauth.init(path) and _load_endpoints()


def _load_endpoints() -> bool:
    global endpoint_data

    if not os.path.isfile(endpoint_data_path()):
        endpoint_data = _get_endpoints()
    else:
        with open(endpoint_data_path()) as ep:
            endpoint_data = json.load(ep)
        if time.time() > endpoint_data[EXP_TIME_KEY]:
            logger.info('Endpoint data expired.')
            endpoint_data = _get_endpoints()

    return True


def _get_endpoints() -> dict:
    global endpoint_data
    r = requests.get(AMZ_ENDPOINT_REQ_URL, headers=oauth.get_auth_header())
    if r.status_code not in OK_CODES:
        logger.critical('Error getting endpoint data. Response: %s' % r.text)
        raise Exception

    try:
        e = r.json()
    except ValueError as e:
        logger.critical('Invalid JSON: "%s"' % r.text)
        raise e

    e[EXP_TIME_KEY] = time.time() + ENDPOINT_VAL_TIME
    endpoint_data = e

    try:
        get_metadata_url()
        get_content_url()
    except KeyError as e:
        logger.critical('Received invalid endpoint data.')
        raise e

    _save_endpoint_data()

    return e


def _save_endpoint_data():
    f = open(endpoint_data_path(), 'w')
    json.dump(endpoint_data, f, indent=4, sort_keys=True)
    f.flush()
    os.fsync(f.fileno())
    f.close()


class RequestError(Exception):
    class CODE(object):
        CONN_EXCEPTION = 1000
        FAILED_SUBREQUEST = 1002
        INCOMPLETE_RESULT = 1003
        REFRESH_FAILED = 1004
        INVALID_TOKEN = 1005

    def __init__(self, status_code: int, msg: str):
        self.status_code = status_code
        if msg:
            self.msg = msg
        else:
            self.msg = '[acd_cli] no body received.'

    def __str__(self):
        return 'RequestError: ' + str(self.status_code) + ', ' + self.msg


def catch_conn_exception(func):
    """Request connection exception decorator
    :raises RequestError"""

    def decorated(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (ConnectionError, ReadTimeoutError) as e:
            raise RequestError(RequestError.CODE.CONN_EXCEPTION, e.__str__())

    return decorated


class BackOffRequest(object):
    """Wrapper for requests that implements timed back-off algorithm
    https://developer.amazon.com/public/apis/experience/cloud-drive/content/best-practices
    Caution: this catches all connection errors and may stall for a long time.
    It is necessary to init this module before use.
    """

    __session = None
    __thr_local = local()
    __lock = Lock()
    __retries = 0
    __next_req = time.time()

    random.seed()

    @classmethod
    def _succeeded(cls):
        with cls.__lock:
            cls.__retries = 0
        cls.__calc_next()

    @classmethod
    def _failed(cls):
        with cls.__lock:
            cls.__retries += 1
        cls.__calc_next()

    @classmethod
    def __calc_next(cls):
        """Calculates minimal acceptable time for next request.
        Back-off time is in a range of seconds, depending on number of failed previous tries (r):
        [0,2^r], maximum interval [0,256]"""
        with cls.__lock:
            duration = random.random() * 2 ** min(cls.__retries, 8)
            cls.__next_req = time.time() + duration

    @classmethod
    def _wait(cls):
        with cls.__lock:
            duration = cls.__next_req - time.time()
        if duration > 5:
            logger.warning('Waiting %fs because of error(s).' % duration)
        logger.debug('Retry %i, waiting %f secs' % (cls.__retries, duration))
        if duration > 0:
            sleep(duration)

    @classmethod
    @catch_conn_exception
    def _request(cls, type_, url: str, acc_codes: list, **kwargs) -> requests.Response:
        if not cls.__session:
            cls.__session = requests.session()
        cls._wait()

        with cls.__lock:
            headers = oauth.get_auth_header()
        if 'headers' in kwargs:
            headers = dict(headers, **(kwargs['headers']))
            del kwargs['headers']

        last_url = getattr(cls.__thr_local, 'last_req_url', None)
        if url == last_url:
            logger.debug('%s "%s"' % (type_, url))
        else:
            logger.info('%s "%s"' % (type_, url))
        if 'data' in kwargs.keys():
            logger.debug(kwargs['data'])

        cls.__thr_local.last_req_url = url

        if 'timeout' in kwargs:
            timeout = kwargs['timeout']
            del kwargs['timeout']
        else:
            timeout = REQUESTS_TIMEOUT

        try:
            r = cls.__session.request(type_, url, headers=headers, timeout=timeout, **kwargs)
        except:
            cls._failed()
            raise

        cls._succeeded() if r.status_code in acc_codes else cls._failed()
        return r

    @classmethod
    def get(cls, url, acc_codes=OK_CODES, **kwargs) -> requests.Response:
        return cls._request('GET', url, acc_codes, **kwargs)

    @classmethod
    def post(cls, url, acc_codes=OK_CODES, **kwargs) -> requests.Response:
        return cls._request('POST', url, acc_codes, **kwargs)

    @classmethod
    def patch(cls, url, acc_codes=OK_CODES, **kwargs) -> requests.Response:
        return cls._request('PATCH', url, acc_codes, **kwargs)

    @classmethod
    def put(cls, url, acc_codes=OK_CODES, **kwargs) -> requests.Response:
        return cls._request('PUT', url, acc_codes, **kwargs)

    @classmethod
    def delete(cls, url, acc_codes=OK_CODES, **kwargs) -> requests.Response:
        return cls._request('DELETE', url, acc_codes, **kwargs)

    @classmethod
    def paginated_get(cls, url: str, params: dict=None) -> list:
        if params is None:
            params = {}
        node_list = []

        while True:
            r = cls.get(url, params=params)
            if r.status_code not in OK_CODES:
                logger.error("Error getting node list.")
                raise RequestError(r.status_code, r.text)
            ret = r.json()
            node_list.extend(ret['data'])
            if 'nextToken' in ret.keys():
                params['startToken'] = ret['nextToken']
            else:
                if ret['count'] != len(node_list):
                    logger.warning(
                        'Expected {} items, received {}.'.format(ret['count'], len(node_list)))
                break

        return node_list



"""Formatters for query bundle iterables"""


import os
import sys

colors = filter(None, os.environ.get('LS_COLORS', '').split(':'))
colors = dict(c.split('=') for c in colors)
# colors is now a mapping of 'type': 'color code' or '*.ext' : 'color code'

seq_tpl = '\x1B[%sm'
res = seq_tpl % colors.get('rs', '')  # reset code
dir_fmt = seq_tpl % colors.get('di', '') + '%s' + res  # dir text
nor_fmt = seq_tpl % colors.get('no', '') + '%s' + res  # 'normal' colored text

ColorMode = dict(auto=0, always=1, never=2)


def init(color=ColorMode['auto']):
    if color == ColorMode['never'] \
            or not res \
            or (color == ColorMode['auto'] and not sys.__stdout__.isatty()):
        global get_adfixes, color_path, color_status, nor_fmt
        get_adfixes = lambda _: ('', '')
        color_path = lambda x: x
        color_status = lambda x: x[0]
        nor_fmt = '%s'


def color_file(name: str) -> str:
    parts = name.split('.')
    if len(parts) > 1:
        ext = parts.pop()
        code = colors.get('*.' + ext)
        if code:
            return seq_tpl % code + name + res

    return nor_fmt % name


def color_path(path: str) -> str:
    segments = path.split('/')
    path_segments = [dir_fmt % s for s in segments[:-1]]
    last_seg = segments[-1] if segments[-1:] else ''
    file_seg = color_file(last_seg)
    return '/'.join(path_segments + [file_seg])


def color_status(status):
    if status == 'AVAILABLE':
        return seq_tpl % '32' + status[0] + res  # green
    if status == 'TRASH':
        return seq_tpl % '31' + status[0] + res  # red


class ListFormatter(object):
    @staticmethod
    def __new__(cls, bunches, **kwargs):
        return LSFormatter(bunches, **kwargs)


class LSFormatter(ListFormatter):
    @staticmethod
    def __new__(cls, bunches, recursive=False):
        is_first = True
        for bunch in bunches:
            node = bunch.node
            children = 0 if not node.is_folder() else len(node.children)
            if bunch.path is None:
                bunch.path = node.containing_folder()
            if recursive and node.is_folder() and not is_first and children > 0:
                yield ''
            yield '[{}] [{}] {}{}'.format(
                nor_fmt % node.id,
                color_status(node.status),
                color_path(bunch.path) if node.is_folder() and children else '',
                color_path(node.simple_name())
            )
            is_first = False


class LongIDFormatter(ListFormatter):
    @staticmethod
    def __new__(cls, bunches):
        for bunch in bunches:
            node = bunch.node
            if bunch.path is None:
                bunch.path = node.containing_folder()
            yield '[{}] [{}] {}{}'.format(
                nor_fmt % node.id,
                color_status(node.status),
                color_path(bunch.path),
                color_path(node.simple_name())
            )

class TreeFormatter(ListFormatter):
    @staticmethod
    def __new__(cls, bunches):
        prev = None
        for bunch in bunches:
            pre = ''
            if bunch.depth > 0:
                pre = ' ' * 4 * bunch.depth
            yield pre + color_path(bunch.node.simple_name())
            prev = bunch



class IDFormatter(ListFormatter):
    @staticmethod
    def __new__(cls, bunches):
        for bunch in bunches:
            yield bunch.node.id


class PathFormatter(ListFormatter):
    @staticmethod
    def __new__(cls, bunches):
        for bunch in bunches:
            yield bunch.node.full_path()



import time
import logging
import queue
from threading import Thread, Event, Lock

from . import progress

_logger = logging.getLogger(__name__)


class QueuedLoader(object):
    """Multi-threaded loader intended for file transfer jobs."""

    MAX_NUM_WORKERS = 8
    MAX_RETRIES = 4
    REFRESH_PROGRESS_INT = 0.3

    def __init__(self, workers=1, print_progress=True, max_retries=0):
        self.workers = min(abs(workers), self.MAX_NUM_WORKERS)
        self.q = queue.Queue()
        self.halt = False
        self.exit_stat = 0
        self.stat_lock = Lock()
        self.print_progress = print_progress
        self.retries = min(abs(max_retries), self.MAX_RETRIES)

        self.mp = progress.MultiProgress()

    def _print_prog(self):
        while not self.halt:
            self.mp.print_progress()
            time.sleep(self.REFRESH_PROGRESS_INT)
        self.mp.end()

    def _worker_task(self, num: int):
        while True:
            try_ = 0
            f = self.q.get()
            while try_ <= self.retries:
                rr = f()
                if not rr.retry:
                    break
                try_ += 1

            with self.stat_lock:
                self.exit_stat |= rr.ret_val
            self.q.task_done()

    def add_jobs(self, jobs: list):
        """:param jobs: list of partials that return a RetryRetVal and have a pg_handler kwarg"""
        for job in jobs:
            h = job.keywords.get('pg_handler')
            self.mp.add(h)
            self.q.put(job)

    def start(self) -> int:
        """Starts worker threads and, if applicable, progress printer thread.
        :returns: accumulated return value"""

        _logger.info('%d jobs in queue.' % self.q.qsize())

        p = None
        print_progress = self.print_progress and self.q.qsize() > 0
        if print_progress:
            p = Thread(target=self._print_prog)
            p.daemon = True
            p.start()

        for i in range(self.workers):
            t = Thread(target=self._worker_task, args=(i,), name='worker-' + str(i))
            t.daemon = True
            t.start()

        self.q.join()
        self.halt = True
        if p:
            p.join()

        return self.exit_stat



import time
import sys
from math import floor, log10
from collections import deque


class FileProgress(object):
    __slots__ = ('current', 'status', 'total')

    def __init__(self, total_sz: int, current: int=0):
        self.total = total_sz
        self.current = current
        self.status = None

    def update(self, chunk):
        self.current += chunk.__sizeof__()

    def reset(self):
        self.current = 0

    def done(self):
        self.current = self.total


class MultiProgress(object):
    """Container that accumulates multiple FileProgress objects"""

    def __init__(self):
        self._progresses = []
        self._last_inv = None
        self._last_prog = 0
        self._last_speeds = deque([0] * 10, 10)

    def end(self):
        self.print_progress()
        print()
        failed = sum(1 for s in self._progresses if s.status)
        if failed:
            print('%d file(s) failed.' % failed)

    def add(self, progress: FileProgress):
        self._progresses.append(progress)

    def print_progress(self):
        total = 0
        current = 0
        complete = 0
        for p in self._progresses:
            total += p.total
            current += p.current
            if p.total <= p.current:
                complete += 1

        self._print(total, current, len(self._progresses), complete)

    def _print(self, total_sz: int, current_sz: int, total_items: int, done: int):
        if not self._last_inv:
            self._last_inv = time.time()

        t = time.time()
        duration = t - self._last_inv
        speed = (current_sz - self._last_prog) / duration if duration else 0
        rate = float(current_sz) / total_sz if total_sz else 1
        self._last_speeds.append(speed)

        avg_speed = float(sum(self._last_speeds)) / len(self._last_speeds)

        self._last_inv, self._last_prog = t, current_sz

        percentage = round(rate * 100, ndigits=2)
        completed = "#" * int(percentage / 3)
        spaces = " " * (33 - len(completed))
        item_width = floor(log10(total_items))
        sys.stdout.write('[%s%s] %s%% of %s  %s/%d %s\r'
                         % (completed, spaces, ('%3.1f' % percentage).rjust(5),
                            (file_size_str(total_sz)).rjust(7), str(done).rjust(item_width + 1), total_items,
                            (speed_str(avg_speed)).rjust(10)))
        sys.stdout.flush()


def speed_str(num: int, suffix='B', time_suffix='/s') -> str:
    for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1000.0:
            return "%3.1f%s%s%s" % (num, unit, suffix, time_suffix)
        num /= 1000.0
    return "%.1f%s%s%s" % (num, 'Y', suffix, time_suffix)


def file_size_str(num: int, suffix='B') -> str:
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.0f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


import os
import json
import requests
import time
import logging
import webbrowser
import datetime
from urllib.parse import urlparse, parse_qs

__all__ = ('init', 'get_auth_header')

logger = logging.getLogger(__name__)
handler = None
""":type: OAuthHandler"""


def init(path: str='') -> bool:
    global handler
    handler = _create_handler(path)
    return True


def get_auth_header() -> dict:
    return {'Authorization': handler.get_auth_token()}


TOKEN_INFO_URL = 'https://api.amazon.com/auth/o2/tokeninfo'

def get_access_token_info() -> dict:
    """json keywords
    int exp: expiration time in sec
    str aud: client id
    user_id, app_id, iat (exp time)
    """
    r = requests.get(TOKEN_INFO_URL, params = {'access_token': handler.oauth_data['access_token']})
    return r.json()


def _create_handler(path: str):
    try:
        return LocalOAuthHandler(path)
    except:
        return AppspotOAuthHandler(path)


class OAuthHandler(object):
    OAUTH_DATA_FILE = 'oauth_data'

    class KEYS(object):
        EXP_IN = 'expires_in'
        ACC_TOKEN = 'access_token'
        REFR_TOKEN = 'refresh_token'
        EXP_TIME = 'exp_time'  # manually added
        REDIRECT_URI = 'redirect_uri' # only for local

    def __init__(self, path):
        self.path = path
        self.oauth_data = {}
        self.oauth_data_path = os.path.join(path, self.OAUTH_DATA_FILE)

        self.exp_time = lambda: self.oauth_data[self.KEYS.EXP_TIME]
        self.init_time = time.time()

    @classmethod
    def validate(cls, oauth: str) -> dict:
        """Validate and deserialize OAuth string
        :raises: RequestError"""

        from .common import RequestError

        try:
            o = json.loads(oauth)
            o[cls.KEYS.ACC_TOKEN]
            o[cls.KEYS.EXP_IN]
            o[cls.KEYS.REFR_TOKEN]
            return o
        except (ValueError, KeyError) as e:
            logger.critical('Invalid authentication token: Invalid JSON or missing key.')
            raise RequestError(RequestError.CODE.INVALID_TOKEN, e.__str__())

    def treat_auth_token(self, time: float):
        """Add expiration time to member OAuth dict using specified begin time."""
        exp_time = time + self.oauth_data[self.KEYS.EXP_IN] - 120
        self.oauth_data[self.KEYS.EXP_TIME] = exp_time
        logger.info('New token expires at %s.'
                    % datetime.datetime.fromtimestamp(exp_time).isoformat(' '))

    def load_oauth_data(self):
        """Load oauth data file, validate and add expiration time if necessary"""
        self.check_oauth_file_exists()

        with open(self.oauth_data_path) as oa:
            o = oa.read()
        self.oauth_data = self.validate(o)
        if self.KEYS.EXP_TIME not in self.oauth_data:
            self.treat_auth_token(self.init_time)
            self.write_oauth_data()
        else:
            self.get_auth_token(reload=False)

    def get_auth_token(self, reload=True) -> str:
        """Get current access token, refreshes if necessary"""
        if time.time() > self.exp_time():
            logger.info('Token expired at %s.'
                        % datetime.datetime.fromtimestamp(self.exp_time()).isoformat(' '))

            # if multiple instances are running, check for updated file
            if reload:
                with open(self.oauth_data_path) as oa:
                    o = oa.read()
                oauth_data = self.validate(o)

            if time.time() > self.exp_time():
                self.refresh_auth_token()
            else:
                logger.info('Externally updated token found in oauth file.')
        return "Bearer " + self.oauth_data[self.KEYS.ACC_TOKEN]

    def write_oauth_data(self):
        """Dump (treated) OAuth dict to file as JSON."""

        f = open(self.oauth_data_path, 'w')
        json.dump(self.oauth_data, f, indent=4, sort_keys=True)
        f.flush()
        os.fsync(f.fileno())
        f.close()

    def refresh_auth_token(self):
        """Fetch a new access token using the refresh token."""
        raise NotImplementedError

    def check_oauth_file_exists(self):
        """Check for OAuth file existence and one-time initialize if necessary. Throws on error."""
        raise NotImplementedError

class AppspotOAuthHandler(OAuthHandler):
    APPSPOT_URL = 'https://tensile-runway-92512.appspot.com/'

    def __init__(self, path):
        super().__init__(path)
        self.load_oauth_data()

        logger.info('%s initialized' % self.__class__.__name__)

    def check_oauth_file_exists(self):
        """:raises Exception"""
        if not os.path.isfile(self.oauth_data_path):
            webbrowser.open_new_tab(AppspotOAuthHandler.APPSPOT_URL)
            input('A browser tab will have/be opened at %s.\nPlease accept the request '
                  % AppspotOAuthHandler.APPSPOT_URL +
                  'and save the plaintext response data into a file called "%s"' % self.OAUTH_DATA_FILE +
                  ' in the directory "%s".\nThen, press a key to continue.\n' % self.path)

            if not os.path.isfile(self.oauth_data_path):
                logger.error('File "%s" not found.' % self.OAUTH_DATA_FILE)
                raise Exception

    def refresh_auth_token(self):
        """:raises RequestError"""

        logger.info('Refreshing authentication token.')

        ref = {self.KEYS.REFR_TOKEN: self.oauth_data[self.KEYS.REFR_TOKEN]}
        t = time.time()

        from .common import RequestError

        try:
            response = requests.post(self.APPSPOT_URL, data=ref)
        except ConnectionError as e:
            logger.critical('Error refreshing authentication token.')
            raise RequestError(RequestError.CODE.CONN_EXCEPTION, e.__str__())

        if response.status_code != requests.codes.ok:
            raise RequestError(RequestError.CODE.REFRESH_FAILED,
                               'Error refreshing authentication token: %s' % response.text)

        r = self.validate(response.text)

        self.oauth_data = r
        self.treat_auth_token(t)
        self.write_oauth_data()


class LocalOAuthHandler(OAuthHandler):
    """https://developer.amazon.com/public/apis/experience/cloud-drive/content/getting-started"""

    CLIENT_DATA_FILE = 'client_data'

    AMAZON_OA_LOGIN_URL = 'https://amazon.com/ap/oa'
    AMAZON_OA_TOKEN_URL = 'https://api.amazon.com/auth/o2/token'
    REDIRECT_URI = 'http://localhost'

    def __init__(self, path):
        """This is intended to raise if the client data file is not found."""

        super().__init__(path)

        self.client_data = {}

        self.client_id = lambda: self.client_data.get('CLIENT_ID')
        self.client_secret = lambda: self.client_data.get('CLIENT_SECRET')

        self.OAUTH_ST1 = lambda: {'client_id': self.client_id(),
                                  'response_type': 'code',
                                  'scope': 'clouddrive:read clouddrive:write',
                                  'redirect_uri': self.REDIRECT_URI}

        self.OAUTH_ST2 = lambda: {'grant_type': 'authorization_code',
                                  'code': None,
                                  'client_id': self.client_id(),
                                  'client_secret': self.client_secret(),
                                  'redirect_uri': self.REDIRECT_URI}

        self.OAUTH_REF = lambda: {'grant_type': 'refresh_token',
                                  'refresh_token': None,
                                  'client_id': self.client_id(),
                                  'client_secret': self.client_secret(),
                                  'redirect_uri': self.REDIRECT_URI}

        self.load_client_data()
        self.load_oauth_data()

        logger.info('%s initialized.' % self.__class__.__name__)

    def load_client_data(self):
        cdp = os.path.join(self.path, self.CLIENT_DATA_FILE)
        with open(cdp) as cd:
            self.client_data = json.load(cd)

        if self.client_id() == '' or self.client_secret() == '':
            logger.critical('Client ID or client secret empty or key absent.')
            raise Exception

    def check_oauth_file_exists(self):
        """:raises Exception"""
        if not os.path.isfile(self.oauth_data_path):
            r = requests.post(self.AMAZON_OA_LOGIN_URL, params=self.OAUTH_ST1())
            if r.status_code != requests.status_codes.codes.ok:
                logging.critical('Error')
                raise Exception

            webbrowser.open_new_tab(r.url)
            print('A window will have opened at %s' % self.AMAZON_OA_LOGIN_URL)

            ret_url = input("Please log in or accept and enter the url you have been redirected to: ")
            ret_q = parse_qs(urlparse(ret_url).query)

            st2 = self.OAUTH_ST2()
            st2['code'] = ret_q['code'][0]

            response = requests.post(self.AMAZON_OA_TOKEN_URL, data=st2)
            self.oauth_data = self.validate(response.text)
            self.write_oauth_data()

    def refresh_auth_token(self):
        logger.info('Refreshing authentication token.')

        ref = self.OAUTH_REF()
        ref[self.KEYS.REFR_TOKEN] = self.oauth_data[self.KEYS.REFR_TOKEN]

        from .common import RequestError

        t = time.time()
        try:
            response = requests.post(self.AMAZON_OA_TOKEN_URL, data=ref)
        except ConnectionError as e:
            logger.critical('Error refreshing authentication token.')
            raise RequestError(RequestError.CODE.CONN_EXCEPTION, e.__str__())

        if response.status_code != requests.codes.ok:
            raise RequestError(RequestError.CODE.REFRESH_FAILED,
                               'Error refreshing authentication token: %s' % response.text)

        self.oauth_data = self.validate(response.text)
        self.treat_auth_token(t)
        self.write_oauth_data()



from .common import *


def list_trash() -> list:
    """retrieves top-level trash list"""
    return BackOffRequest.paginated_get(get_metadata_url() + 'trash')


def move_to_trash(node_id: str) -> dict:
    r = BackOffRequest.put(get_metadata_url() + 'trash/' + node_id)
    if r.status_code not in OK_CODES:
        raise RequestError(r.status_code, r.text)
    return r.json()


def restore(node_id: str) -> dict:
    r = BackOffRequest.post(get_metadata_url() + 'trash/' + node_id + '/restore')
    if r.status_code not in OK_CODES:
        raise RequestError(r.status_code, r.text)
    return r.json()


# {"message":"Insufficient permissions granted for operation: purgeNode"}
def purge(node_id: str) -> dict:
    r = BackOffRequest.delete(get_metadata_url() + 'nodes/' + node_id)
    if r.status_code not in OK_CODES:
        raise RequestError(r.status_code, r.text)
    return r.json()


import os
import logging
from sqlalchemy import *
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session, relationship
from sqlalchemy.exc import DatabaseError
from datetime import datetime, timedelta

logger = logging.getLogger(__name__)

Session = None
engine = None

Base = declarative_base()

DB_SCHEMA_VER = 1
DB_FILENAME = 'nodes.db'

parentage_table = Table('parentage', Base.metadata,
                        Column('parent', String(50), ForeignKey('folders.id'), primary_key=True),
                        Column('child', String(50), ForeignKey('nodes.id'), primary_key=True)
                        )


class Metadate(Base):
    """added in v1"""
    __tablename__ = 'metadata'

    key = Column(String(64), primary_key=True)
    value = Column(String)

    def __init__(self, key: str, value: str):
        self.key = key
        self.value = value


class _KeyValueStorage(object):
    @staticmethod
    def __getitem__(key: str):
        val = Session.query(Metadate).filter_by(key=key).first()
        if val:
            return val.value
        else:
            raise KeyError

    @staticmethod
    def __setitem__(key: str, value: str):
        md = Metadate(key, value)
        Session.merge(md)
        Session.commit()

    @staticmethod
    def __len__():
        return Session.query(Metadate).count()

    @staticmethod
    def get(key: str, default: str=None):
        val = Session.query(Metadate).filter_by(key=key).first()
        return val.value if val else default

    @classmethod
    def update(cls, dict_: dict):
        for key in dict_.keys():
            cls.__setitem__(key, dict_[key])

# next best thing to a subscriptable class
KeyValueStorage = _KeyValueStorage()


class Label(Base):
    """added in v1"""
    __tablename__ = 'labels'

    id = Column(String(50), ForeignKey('nodes.id'), primary_key=True)
    name = Column(String(256), primary_key=True)


# TODO: cycle safety for full_path()
class Node(Base):
    __tablename__ = 'nodes'

    # apparently 22 chars; max length is 22 for folders according to API doc ?!
    id = Column(String(50), primary_key=True, unique=True)
    type = Column(String(15))
    name = Column(String(256))
    description = Column(String(500))

    created = Column(DateTime)
    modified = Column(DateTime)
    updated = Column(DateTime)

    # "pending" status seems to be reserved for not yet finished uploads
    status = Column(Enum('AVAILABLE', 'TRASH', 'PURGED', 'PENDING'))

    __mapper_args__ = {
        'polymorphic_identity': 'node',
        'polymorphic_on': type
    }

    def __lt__(self, other):
        """Compares this node to another one on same path level. Sorts case-sensitive, Folder first. """
        if isinstance(self, Folder):
            if isinstance(other, File):
                return True
            return self.name < other.name
        if isinstance(other, Folder):
            return False
        return self.name < other.name

    def is_file(self) -> bool:
        return isinstance(self, File)

    def is_folder(self) -> bool:
        return isinstance(self, Folder)

    def is_available(self) -> bool:
        return self.status == 'AVAILABLE'

    def id_str(self) -> str:
        """short id string containing id, stat, name"""
        return '[{}] [{}] {}'.format(self.id, self.status[0], self.simple_name())

    def containing_folder(self) -> str:
        if len(self.parents) == 0:
            return ''
        return self.parents[0].full_path()

    parents = relationship('Folder', secondary=parentage_table,
                           primaryjoin=id == parentage_table.c.child,
                           backref='children'
                           )


class File(Node):
    __tablename__ = 'files'

    id = Column(String(50), ForeignKey('nodes.id'), primary_key=True, unique=True)
    md5 = Column(String(32))
    size = Column(BigInteger)

    __mapper_args__ = {
        'polymorphic_identity': 'file'
    }

    def __init__(self, id: str, name: str, created: datetime, modified: datetime, md5: str, size: int, status: Enum):
        self.id = id
        self.name = name
        self.created = created.replace(tzinfo=None)
        self.modified = modified.replace(tzinfo=None)
        self.md5 = md5
        self.size = size
        self.status = status

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return (self.id == other.id and self.name == other.name
                and self.created - other.created == timedelta(0)
                and self.modified - other.modified == timedelta(0)
                and self.md5 == other.md5 and self.size == other.size
                and self.status == other.status)

    def __repr__(self):
        return 'File(%r, %r)' % (self.id, self.name)

    def simple_name(self) -> str:
        """file name"""
        return self.name

    def full_path(self) -> str:
        """absolute path of file (first containing folder chain)"""
        if len(self.parents) == 0:
            return self.name
        return self.parents[0].full_path() + self.name


class Folder(Node):
    __tablename__ = 'folders'

    id = Column(String(50), ForeignKey('nodes.id'), primary_key=True, unique=True)

    __mapper_args__ = {
        'polymorphic_identity': 'folder'
    }

    # noinspection PyShadowingBuiltins
    def __init__(self, id: str, name: str, created: datetime, modified: datetime, status: Enum):
        self.id = id
        self.name = name
        self.created = created.replace(tzinfo=None)
        self.modified = modified.replace(tzinfo=None)
        self.status = status

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return (self.id == other.id
                and self.name == other.name
                and self.created == other.created
                and self.modified == other.modified
                and self.status == other.status)

    def __repr__(self):
        return 'Folder(%r, %r)' % (self.id, self.name)

    def simple_name(self) -> str:
        return (self.name if self.name else '') + '/'

    # path of first occurrence
    def full_path(self) -> str:
        if len(self.parents) == 0:
            return '/'
        return self.parents[0].full_path() + self.simple_name()

    def get_child(self, name: str) -> Node:
        """ Gets non-trashed child by name. """
        for child in self.children:
            if child.name == name and child.status != 'TRASH':
                return child
        return

    def available_children(self):
        for c in self.children:
            if c.is_available():
                yield c


"""End of 'schema'"""


def init(path=''):
    logger.info('Initializing cache with path "%s".' % os.path.realpath(path))
    db_path = os.path.join(path, DB_FILENAME)

    # doesn't seem to work on Windows
    from ctypes import util, CDLL

    try:
        lib = util.find_library('sqlite3')
    except OSError:
        logger.info('Skipping sqlite thread-safety test.')
    else:
        if lib:
            dll = CDLL(lib)
            if dll and not dll.sqlite3_threadsafe():
                # http://www.sqlite.org/c3ref/threadsafe.html
                logger.warning('Your sqlite3 version was compiled without mutexes. It is not thread-safe.')

    global Session
    global engine
    engine = create_engine('sqlite:///%s' % db_path, connect_args={'check_same_thread': False})

    # check for serialized mode

    uninitialized = not os.path.exists(db_path)
    if not uninitialized:
        try:
            uninitialized = not engine.has_table(Metadate.__tablename__) and \
                            not engine.has_table(Node.__tablename__) and \
                            not engine.has_table(File.__tablename__) and \
                            not engine.has_table(Folder.__tablename__)
        except DatabaseError:
            logger.critical('Error opening database.')
            return False

    integrity_check()

    if uninitialized:
        r = engine.execute('PRAGMA user_version = %i;' % DB_SCHEMA_VER)
        r.close()

    logger.info('Cache %sconsidered uninitialized.' % ('' if uninitialized else 'not '))

    Base.metadata.create_all(engine)
    session_factory = sessionmaker(bind=engine)
    Session = scoped_session(session_factory)

    if uninitialized:
        return True

    r = engine.execute('PRAGMA user_version;')
    ver = r.first()[0]
    r.close()

    logger.info('DB schema version is %s.' % ver)

    if DB_SCHEMA_VER > ver:
        _migrate(ver)

    return True


def integrity_check():
    r = engine.execute('PRAGMA integrity_check;')
    if r.first()[0] != 'ok':
        logger.warn('Sqlite database integrity check failed. '
                    'You may need to clear the cache if you encounter any errors.')


def dump_table_sql():
    def dump(sql, *multiparams, **params):
        print(sql.compile(dialect=engine.dialect))
    engine = create_engine('sqlite://', strategy='mock', executor=dump)
    Base.metadata.create_all(engine, checkfirst=False)


def drop_all():
    Base.metadata.drop_all(engine)
    logger.info('Dropped all tables.')


def remove_db_file(path: str):
    db_path = os.path.join(path, DB_FILENAME)
    try:
        os.remove(db_path)
    except OSError:
        logger.critical('Error removing database file "%s".' % db_path)


def _migrate(schema: int):
    """ Migrate database to highest schema
    :param schema: current (cache file) schema version to upgrade from
    """

    migrations = [_0_to_1]

    conn = engine.connect()
    trans = conn.begin()
    try:
        for update_ in migrations[schema:]:
            logger.warning('Updating db schema from %i to %i.' % (schema, schema + 1))
            update_(conn)
            schema += 1
        trans.commit()
    except:
        trans.rollback()
        raise

    conn.close()


def _0_to_1(conn):
    conn.execute('ALTER TABLE nodes ADD updated DATETIME;')
    conn.execute('ALTER TABLE nodes ADD description VARCHAR(500);')
    conn.execute('PRAGMA user_version = 1;')



import logging
from sqlalchemy import func

from . import db

logger = logging.getLogger(__name__)


class Bunch:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __str__(self):
        return str(self.__dict__)


def get_node(node_id: str) -> db.Node:
    return db.Session.query(db.Node).filter_by(id=node_id).first()


# may be broken
def get_root_node() -> db.Folder:
    return db.Session.query(db.Folder).filter_by(name=None).first()


def get_root_id() -> str:
    root = get_root_node()
    if root:
        return root.id


def get_node_count() -> int:
    return db.Session.query(db.Node).count()


def get_file_count() -> int:
    return db.Session.query(db.File).count()


def calculate_usage() -> int:
    return db.Session.query(func.sum(db.File.size)).scalar()


def file_size(id: str) -> int:
    return db.Session.query(db.File).filter_by(id=id).first().size


def tree(root_id: str=None, trash=False):
    if root_id is None:
        return walk_nodes(trash=trash)

    folder = db.Session.query(db.Folder).filter_by(id=root_id).first()
    if not folder:
        logger.error('Not a folder or not found: "%s".' % root_id)
        return

    return walk_nodes(folder, True, True, trash)


def list_children(folder_id: str, recursive=False, trash=False):
    """ Creates Bunches of folder's children
    :param folder_id: valid folder's id
    """
    folder = db.Session.query(db.Folder).filter_by(id=folder_id).first()
    if not folder:
        logger.warning('Not a folder or not found: "%s".' % folder_id)
        return

    return walk_nodes(folder, False, recursive, trash)


# TODO: refashion this to return a tuple like os.walk
def walk_nodes(root: db.Folder=None, add_root=True, recursive=True, trash=False, path='', depth=0):
    """ Generates Bunches of (non-)trashed nodes
    :param root: start folder
    :param add_root: whether to add the (uppermost) root node and prepend its path to its children
    :param recursive: whether to traverse hierarchy
    :param trash: whether to include trash
    :param path: the path on which this method incarnation was reached
    :rtype: Iterable[Bunch]
    :return: list of Bunches including node and path attributes
    """

    if not root:
        root = get_root_node()
        if not root:
            return

    if add_root:
        yield Bunch(node=root, path=path, depth=depth)
        path += root.simple_name()

    if not recursive:
        children = sorted(root.children)
    else:
        children = sorted(root.children, key=lambda x: ('b' if x.is_folder() else 'a') + x.name)

    for child in children:
        if child.status == 'TRASH' and not trash:
            continue
        if isinstance(child, db.Folder) and recursive:
            for node in walk_nodes(child, True, recursive, trash, path, depth + 1):
                yield node
        else:
            yield Bunch(node=child, path=path, depth=depth + 1)


def list_trash(recursive=False):
    trash_nodes = db.Session().query(db.Node).filter(db.Node.status == 'TRASH').all()
    trash_nodes = sorted(trash_nodes)

    for node in trash_nodes:
        yield Bunch(node=node, path=node.containing_folder())
        if isinstance(node, db.Folder) and recursive:
            for child in walk_nodes(node, False, True, True, node.full_path()):
                yield child


def find(name: str):
    q = db.Session.query(db.Node).filter(db.Node.name.like('%' + name + '%'))
    q = sorted(q, key=lambda x: x.full_path())

    for node in q:
        yield Bunch(node=node, path=node.containing_folder())


def find_md5(md5: str):
    q = db.Session.query(db.File).filter_by(md5=md5)
    q = sorted(q, key=lambda x: x.full_path())

    for node in q:
        yield Bunch(node=node, path=node.containing_folder())


def file_size_exists(size: int) -> bool:
    """Returns whether cache contains one or more file(s) of given size."""
    return db.Session.query(db.File).filter_by(size=size).count()


def resolve_path(path: str, trash=True) -> tuple:
    node, _ = resolve(path, None, trash)
    return node.id if node else None


def resolve(path: str, root=None, trash=True) -> tuple:
    """Resolves absolute path to (node, parent) tuple if fully unique"""
    if not path or (not root and '/' not in path):
        return None, None

    segments = path.split('/')
    if segments[0] == '' and not root:
        root = get_root_node()
        # empty cache
        if not root:
            return None, None

    if not root:
        return None, None

    if len(segments) == 1 or segments[1] == '':
        return root, None

    segments = segments[1:]

    # TODO
    if root.is_file() or (not root.is_available() and not trash):
        return None, None

    children = []  # possibly non-unique trash children
    for child in root.children:
        if child.name == segments[0]:
            if child.is_available():
                n, p = resolve('/'.join(segments), child, trash)
                return n, (p if p else root)
            children.append(child)

    if not trash:
        return None, None
    ids = []
    for trash_child in children:
        res, _ = resolve('/'.join(segments), trash_child)
        if res:
            ids.append(res)
    if len(ids) == 1:
        return ids[0], root
    elif len(ids) == 0:
        logger.debug('Could not resolve path "%s"' % path)
    else:
        logger.info('Could not resolve non fully unique (i.e. trash) path "%s"' % path)

    return None, None



#!/usr/bin/env python3
import sys
import os
import json
import argparse
import logging
import signal
from datetime import datetime, timedelta
import time
import re
import appdirs
from functools import partial
from collections import namedtuple

from pkgutil import walk_packages
from pkg_resources import iter_entry_points

from acdcli.api import account, common, content, metadata, trash
from acdcli.api.common import RequestError
from acdcli.cache import *
from acdcli.utils import hashing, progress
from acdcli.utils.threading import QueuedLoader

# load local plugin modules (default ones, for developers)
from acdcli import plugins

for importer, modname, ispkg in walk_packages(path=plugins.__path__, prefix=plugins.__name__ + '.',
                                              onerror=lambda x: None):
    if not ispkg:
        __import__(modname)

# load additional plugins from entry point
for plug_mod in iter_entry_points(group='acdcli.plugins', name=None):
    __import__(plug_mod.module_name)

__version__ = '0.3.0a1'
_app_name = 'acd_cli'

logger = logging.getLogger(_app_name)

# noinspection PyBroadException
# monkey patch the user agent
try:
    import requests.utils

    requests.utils.old_dau = requests.utils.default_user_agent

    def new_dau():
        return _app_name + '/' + __version__ + ' ' + requests.utils.old_dau()

    requests.utils.default_user_agent = new_dau
except:
    pass

# path settings

cp = os.environ.get('ACD_CLI_CACHE_PATH')
sp = os.environ.get('ACD_CLI_SETTINGS_PATH')

CACHE_PATH = cp if cp else appdirs.user_cache_dir(_app_name)
SETTINGS_PATH = sp if sp else appdirs.user_config_dir(_app_name)

if not os.path.isdir(CACHE_PATH):
    try:
        os.makedirs(CACHE_PATH, mode=0o0700)  # private data
    except OSError:
        logger.critical('Error creating cache directory "%s"' % CACHE_PATH)
        sys.exit(1)

# return values

INVALID_ARG_RETVAL = 2  # doubles as flag
INIT_FAILED_RETVAL = 3
KEYB_INTERR_RETVAL = 4

# additional retval flags
UL_DL_FAILED = 8
UL_TIMEOUT = 16
HASH_MISMATCH = 32
ERR_CR_FOLDER = 64
SIZE_MISMATCH = 128
CACHE_ASYNC = 256
DUPLICATE = 512


def signal_handler(signal_, frame):
    # if db.Session:
    #     db.Session.rollback()
    sys.exit(KEYB_INTERR_RETVAL)


signal.signal(signal.SIGINT, signal_handler)
if hasattr(signal, 'SIGPIPE'):
    signal.signal(signal.SIGPIPE, signal_handler)


def pprint(d: dict):
    print(json.dumps(d, indent=4, sort_keys=True))


#
# Glue functions (API, cache)
#


class CacheConsts(object):
    CHECKPOINT_KEY = 'checkpoint'
    LAST_SYNC_KEY = 'last_sync'
    MAX_AGE = 30


def sync_node_list(full=False):
    cp = db.KeyValueStorage.get(CacheConsts.CHECKPOINT_KEY) if not full else None

    try:
        nodes, purged, ncp, full = metadata.get_changes(checkpoint=cp, include_purged=not not cp)
    except RequestError as e:
        logger.critical('Sync failed.')
        print(e)
        return 1

    if full:
        db.drop_all()
        db.init(CACHE_PATH)
    else:
        sync.remove_purged(purged)

    if len(nodes) > 0:
        sync.insert_nodes(nodes, partial=not full)
    db.KeyValueStorage.update(
        {CacheConsts.CHECKPOINT_KEY: ncp, CacheConsts.LAST_SYNC_KEY: time.time()})
    return


def old_sync():
    db.drop_all()
    db.init(CACHE_PATH)
    try:
        folders = metadata.get_folder_list()
        folders.extend(metadata.get_trashed_folders())
        files = metadata.get_file_list()
        files.extend(metadata.get_trashed_files())
    except RequestError as e:
        logger.critical('Sync failed.')
        print(e)
        return 1

    sync.insert_nodes(files + folders, partial=False)
    db.KeyValueStorage['sync_date'] = time.time()

#
# File transfer
#

RetryRetVal = namedtuple('RetryRetVal', ['ret_val', 'retry'])
RETRY_RETVALS = [UL_DL_FAILED]


def retry_on(ret_vals: list):
    """:param ret_vals: list of retry values on which execution should be repeated"""

    def wrap(f):
        def wrapped(*args, **kwargs):
            ret_val = 1
            try:
                ret_val = f(*args, **kwargs)
            except Exception as e:
                import traceback
                logger.error(traceback.format_exc())
            h = kwargs.get('pg_handler')
            h.status = ret_val
            retry = ret_val in ret_vals
            if retry:
                h.reset()
            return RetryRetVal(ret_val, ret_val in ret_vals)

        return wrapped

    return wrap


def compare_hashes(hash1: str, hash2: str, file_name: str):
    if hash1 != hash2:
        logger.error('Hash mismatch between local and remote file for "%s".' % file_name)
        return HASH_MISMATCH

    logger.debug('Local and remote hashes match for "%s".' % file_name)
    return 0


def create_upload_jobs(path: str, parent_id: str, overwr: bool, force: bool, dedup: bool,
                       exclude: list, jobs: list) -> int:
    if not os.access(path, os.R_OK):
        logger.error('Path "%s" is not accessible.' % path)
        return INVALID_ARG_RETVAL

    if os.path.isdir(path):
        return traverse_ul_dir(path, parent_id, overwr, force, dedup, exclude, jobs)
    elif os.path.isfile(path):
        short_nm = os.path.basename(path)
        for reg in exclude:
            if re.match(reg, short_nm):
                logger.info('Skipping upload of "%s" because of exclusion pattern.' % short_nm)
                return 0

        prog = progress.FileProgress(os.path.getsize(path))
        fo = partial(upload_file, path, parent_id, overwr, force, dedup, pg_handler=prog)
        jobs.append(fo)
        return 0


def traverse_ul_dir(directory: str, parent_id: str, overwr: bool, force: bool, dedup: bool,
                    exclude: list, jobs: list) -> int:
    """Duplicates local directory structure."""

    if parent_id is None:
        parent_id = query.get_root_id()
    parent = query.get_node(parent_id)

    real_path = os.path.realpath(directory)
    short_nm = os.path.basename(real_path)

    curr_node = parent.get_child(short_nm)
    if not curr_node or not curr_node.is_available() or not parent.is_available():
        try:
            r = content.create_folder(short_nm, parent_id)
            logger.info('Created folder "%s"' % (parent.full_path() + short_nm))
            sync.insert_node(r)
            curr_node = query.get_node(r['id'])
        except RequestError as e:
            if e.status_code == 409:
                logger.error('Folder "%s" already exists. Please sync.' % short_nm)
            else:
                logger.error('Error creating remote folder "%s".' % short_nm)
            return ERR_CR_FOLDER
    elif curr_node.is_file():
        logger.error(
            'Cannot create remote folder "%s", because a file of the same name already exists.' % short_nm)
        return ERR_CR_FOLDER

    try:
        entries = sorted(os.listdir(directory))
    except OSError as e:
        logger.error('Skipping directory %s because of an error.' % directory)
        logger.info(e)
        return

    ret_val = 0
    for entry in entries:
        full_path = os.path.join(real_path, entry)
        ret_val |= create_upload_jobs(full_path, curr_node.id, overwr, force, dedup, exclude, jobs)

    return ret_val


@retry_on(RETRY_RETVALS)
def upload_file(path: str, parent_id: str, overwr: bool, force: bool, dedup: bool,
                pg_handler: progress.FileProgress=None) -> RetryRetVal:
    short_nm = os.path.basename(path)

    if dedup and query.file_size_exists(os.path.getsize(path)):
        nodes = query.find_md5(hashing.hash_file(path))
        nodes = [n for n in format.PathFormatter(nodes)]
        if len(nodes) > 0:
            # print('Skipping upload of duplicate file "%s".' % short_nm)
            logger.info('Location of duplicates: %s' % nodes)
            pg_handler.done()
            return DUPLICATE

    cached_file = query.get_node(parent_id).get_child(short_nm)
    file_id = None
    if cached_file:
        file_id = cached_file.id

    if not file_id:
        logger.info('Uploading %s' % path)
        hasher = hashing.IncrementalHasher()
        try:
            r = content.upload_file(path, parent_id,
                                    read_callbacks=[hasher.update, pg_handler.update],
                                    deduplication=dedup)
        except RequestError as e:
            if e.status_code == 409:  # might happen if cache is outdated
                if not dedup:
                    logger.error('Uploading "%s" failed. Name collision with non-cached file. '
                                 'If you want to overwrite, please sync and try again.' % short_nm)
                else:
                    logger.error(
                        'Uploading "%s" failed. Name or hash collision with non-cached file.' % short_nm)
                    logger.info(e)
                # colliding node ID is returned in error message -> could be used to continue
                return CACHE_ASYNC
            elif e.status_code == 504 or e.status_code == 408:  # proxy timeout / request timeout
                logger.warning('Timeout while uploading "%s".' % short_nm)
                # TODO: wait; request parent folder's children
                return UL_TIMEOUT
            else:
                logger.error(
                    'Uploading "%s" failed. Code: %s, msg: %s' % (short_nm, e.status_code, e.msg))
                return UL_DL_FAILED
        else:
            sync.insert_node(r)
            file_id = r['id']
            md5 = query.get_node(file_id).md5
            return compare_hashes(hasher.get_result(), md5, short_nm)

    # else: file exists
    rmod = (cached_file.modified - datetime(1970, 1, 1)) / timedelta(seconds=1)
    rmod = datetime.utcfromtimestamp(rmod)
    lmod = datetime.utcfromtimestamp(os.path.getmtime(path))
    lcre = datetime.utcfromtimestamp(os.path.getctime(path))

    logger.info('Remote mtime: %s, local mtime: %s, local ctime: %s' % (rmod, lmod, lcre))

    if not overwr and not force:
        logger.info('Skipping upload of existing file "%s".' % short_nm)
        pg_handler.done()
        return 0

    # ctime is checked because files can be overwritten by files with older mtime
    if rmod < lmod or (rmod < lcre and cached_file.size != os.path.getsize(path)) \
            or force:
        return overwrite(file_id, path, dedup=dedup, pg_handler=pg_handler).ret_val
    elif not force:
        logger.info('Skipping upload of "%s" because of mtime or ctime and size.' % short_nm)
        pg_handler.done()
        return 0


@retry_on(RETRY_RETVALS)
def overwrite(node_id, local_file, dedup=False,
              pg_handler: progress.FileProgress=None) -> RetryRetVal:
    hasher = hashing.IncrementalHasher()
    try:
        r = content.overwrite_file(node_id, local_file,
                                   read_callbacks=[hasher.update, pg_handler.update],
                                   deduplication=dedup)
        sync.insert_node(r)
        node = query.get_node(r['id'])
        md5 = node.md5

        return compare_hashes(md5, hasher.get_result(), local_file)
    except RequestError as e:
        logger.error('Error overwriting file. Code: %s, msg: %s' % (e.status_code, e.msg))
        return UL_DL_FAILED


def create_dl_jobs(node_id: str, local_path: str, exclude: list, jobs: list) -> int:
    """Populates passed jobs list with download partials."""
    local_path = local_path if local_path else ''

    node = query.get_node(node_id)
    if not node.is_available():
        return 0

    if node.is_folder():
        return traverse_dl_folder(node_id, local_path, exclude, jobs)

    loc_name = node.name

    for reg in exclude:
        if re.match(reg, loc_name):
            logger.info('Skipping download of "%s" because of exclusion pattern.' % loc_name)
            return 0

    flp = os.path.join(local_path, loc_name)
    if os.path.isfile(flp):
        logger.info('Skipping download of existing file "%s"' % loc_name)
        if os.path.getsize(flp) != node.size:
            logger.info('Skipped file "%s" has different size than local file.' % loc_name)
            return SIZE_MISMATCH
        return 0

    prog = progress.FileProgress(node.size)
    fo = partial(download_file, node_id, local_path, pg_handler=prog)
    jobs.append(fo)

    return 0


def traverse_dl_folder(node_id: str, local_path: str, exclude: list, jobs: list) -> int:
    """Duplicates remote folder structure."""

    if not local_path:
        local_path = os.getcwd()

    node = query.get_node(node_id)

    if node.name is None:
        curr_path = os.path.join(local_path, 'acd')
    else:
        curr_path = os.path.join(local_path, node.name)

    try:
        os.makedirs(curr_path, exist_ok=True)
    except OSError:
        logger.error('Error creating directory "%s".' % curr_path)
        return ERR_CR_FOLDER

    ret_val = 0
    children = sorted(node.children)
    for child in children:
        ret_val |= create_dl_jobs(child.id, curr_path, exclude, jobs)
    return ret_val


@retry_on(RETRY_RETVALS)
def download_file(node_id: str, local_path: str,
                  pg_handler: progress.FileProgress=None) -> RetryRetVal:
    node = query.get_node(node_id)
    name, md5, size = node.name, node.md5, node.size
    # db.Session.remove()  # otherwise, sqlalchemy will complain if thread crashes

    logger.info('Downloading "%s"' % name)

    hasher = hashing.IncrementalHasher()
    try:
        content.download_file(node_id, name, local_path, length=size,
                              write_callbacks=[hasher.update, pg_handler.update])
    except RequestError as e:
        logger.error('Downloading "%s" failed. Code: %s, msg: %s' % (name, e.status_code, e.msg))
        return UL_DL_FAILED
    else:
        return compare_hashes(hasher.get_result(), md5, name)


def compare(local: str, remote_id):
    pass

#
# Subparser actions. Return value [typeof(None), int] will be used as sys exit status.
#

nocache_actions = []
offline_actions = []
no_autores_trash_actions = []


def nocache_action(func):
    """Decorator for actions that do not need to read from cache or autoresolve."""
    nocache_actions.append(func)
    return func


def offline_action(func):
    """Decorator for actions that can be performed without API calls."""
    offline_actions.append(func)
    return func


def no_autores_trash_action(func):
    """Decorator for actions that should not have trash paths auto-resolved."""
    no_autores_trash_actions.append(func)
    return func


###

def sync_action(args: argparse.Namespace):
    print('Syncing...')
    r = sync_node_list(full=args.full)
    print('Done.')
    return r


def old_sync_action(args: argparse.Namespace):
    print('Syncing...')
    r = old_sync()
    print('Done.')
    return r


@nocache_action
@offline_action
def delete_everything_action(args: argparse.Namespace):
    from distutils.util import strtobool
    from shutil import rmtree

    a = input('Do you really want to delete %s? [y/n] ' % CACHE_PATH)
    try:
        if strtobool(a):
            rmtree(CACHE_PATH)
    except ValueError:
        pass
    except OSError as e:
        print(e)
        print('Deleting directory failed.')


@nocache_action
@offline_action
def clear_action(args: argparse.Namespace):
    db.remove_db_file(CACHE_PATH)


@nocache_action
@offline_action
def print_version_action(args: argparse.Namespace):
    print(' '.join([_app_name, __version__]))


@offline_action
def tree_action(args: argparse.Namespace):
    tree = query.tree(args.node, args.include_trash)
    tree = format.TreeFormatter(tree)
    for node in tree:
        print(node)


@nocache_action
def usage_action(args: argparse.Namespace):
    r = account.get_account_usage()
    print(r, end='')


@nocache_action
def quota_action(args: argparse.Namespace):
    r = account.get_quota()
    pprint(r)


def regex_helper(args: argparse.Namespace) -> list:
    """Pre-compiles regexes from strings in args namespace."""
    excl_re = []
    for re_ in args.exclude_re:
        try:
            excl_re.append(re.compile(re_, flags=re.IGNORECASE))
        except re.error as e:
            logger.critical('Invalid regular expression: %s. %s' % (re_, e))
            sys.exit(INVALID_ARG_RETVAL)

    for ending in args.exclude_fe:
        excl_re.append(re.compile('^.*\.' + re.escape(ending) + '$', flags=re.IGNORECASE))

    return excl_re


@no_autores_trash_action
def upload_action(args: argparse.Namespace) -> int:
    if not query.get_node(args.parent):
        logger.critical('Invalid upload folder.')
        return INVALID_ARG_RETVAL

    excl_re = regex_helper(args)

    jobs = []
    ret_val = 0
    for path in args.path:
        if not os.path.exists(path):
            logger.error('Path "%s" does not exist.' % path)
            ret_val |= INVALID_ARG_RETVAL
            continue

        ret_val |= create_upload_jobs(path, args.parent, args.overwrite, args.force,
                                      args.deduplicate, excl_re, jobs)

    ql = QueuedLoader(args.max_connections, max_retries=args.max_retries)
    ql.add_jobs(jobs)

    return ret_val | ql.start()


def overwrite_action(args: argparse.Namespace) -> int:
    if not os.path.isfile(args.file):
        logger.error('Invalid file.')
        return INVALID_ARG_RETVAL

    prog = progress.FileProgress(os.path.getsize(args.file))
    ql = QueuedLoader(max_retries=args.max_retries)
    job = partial(overwrite, args.node, args.file, pg_handler=prog)
    ql.add_jobs([job])

    return ql.start()


def download_action(args: argparse.Namespace) -> int:
    excl_re = regex_helper(args)

    jobs = []
    ret_val = 0
    ret_val |= create_dl_jobs(args.node, args.path, excl_re, jobs)

    ql = QueuedLoader(args.max_connections, max_retries=args.max_retries)
    ql.add_jobs(jobs)

    return ret_val | ql.start()


def create_action(args: argparse.Namespace) -> int:
    parent, folder = os.path.split(args.new_folder)
    # no trailing slash
    if not folder:
        parent, folder = os.path.split(parent)

    if not folder:
        logger.error('Cannot create folder with empty name.')
        return INVALID_ARG_RETVAL

    if parent[-1:] == '' or parent[0] != '/':
        parent = '/' + parent
    p_id = query.resolve_path(parent)
    if not p_id:
        logger.error('Invalid parent path "%s".' % parent)
        return INVALID_ARG_RETVAL

    try:
        r = content.create_folder(folder, p_id)
    except RequestError as e:
        logger.debug(str(e.status_code) + e.msg)
        if e.status_code == 409:
            logger.warning('Folder "%s" already exists.' % folder)
        else:
            logger.error('Error creating folder "%s".' % folder)
            return ERR_CR_FOLDER
    else:
        sync.insert_node(r)


@no_autores_trash_action
@offline_action
def list_trash_action(args: argparse.Namespace):
    t_list = query.list_trash(args.recursive)
    for node in format.ListFormatter(t_list, recursive=args.recursive):
        print(node)


def trash_action(args: argparse.Namespace) -> int:
    try:
        r = trash.move_to_trash(args.node)
        sync.insert_node(r)
    except RequestError as e:
        print(e)
        return 1


def restore_action(args: argparse.Namespace) -> int:
    try:
        r = trash.restore(args.node)
    except RequestError as e:
        logger.error('Error restoring "%s"' % args.node, e)
        return 1
    sync.insert_node(r)


@offline_action
def resolve_action(args: argparse.Namespace) -> int:
    node = query.resolve_path(args.path)
    if node:
        print(node)
    else:
        return INVALID_ARG_RETVAL


@offline_action
def find_action(args: argparse.Namespace):
    r = query.find(args.name)
    r = format.LongIDFormatter(r)
    for node in r:
        print(node)
    if not r:
        return INVALID_ARG_RETVAL


@offline_action
def find_md5_action(args: argparse.Namespace) -> int:
    if len(args.md5) != 32:
        logger.critical('Invalid MD5 specified')
        return 2
    nodes = query.find_md5(args.md5)
    for node in format.LongIDFormatter(nodes):
        print(node)


@offline_action
def children_action(args: argparse.Namespace) -> int:
    nodes = query.list_children(args.node, args.recursive, args.include_trash)
    for entry in format.ListFormatter(nodes, recursive=args.recursive):
        print(entry)


def move_action(args: argparse.Namespace) -> int:
    try:
        r = metadata.move_node(args.child, args.parent)
        sync.insert_node(r)
    except RequestError as e:
        print(e)
        return 1


def rename_action(args: argparse.Namespace) -> int:
    try:
        r = metadata.rename_node(args.node, args.name)
        sync.insert_node(r)
    except RequestError as e:
        print(e)
        return 1


def add_child_action(args: argparse.Namespace) -> int:
    try:
        r = metadata.add_child(args.parent, args.child)
        sync.insert_node(r)
    except RequestError as e:
        print(e)
        return 1


def remove_child_action(args: argparse.Namespace) -> int:
    try:
        r = metadata.remove_child(args.parent, args.child)
        sync.insert_node(r)
    except RequestError as e:
        print(e)
        return 1


def metadata_action(args: argparse.Namespace) -> int:
    try:
        r = metadata.get_metadata(args.node)
        pprint(r)
    except RequestError as e:
        print(e)
        return INVALID_ARG_RETVAL


@offline_action
@nocache_action
def dump_sql_action(args: argparse.Namespace):
    db.dump_table_sql()


def mount_action(args: argparse.Namespace):
    import acdcli.fuse
    acdcli.fuse.mount(args.path, ro=args.ro, foreground=args.foreground)


@offline_action
def compare_action(args: argparse.Namespace):
    pass


#
# helper methods
#


# added for version 0.1.3 on 15-05-04
def migrate_cache_files():
    files = ['oauth_data', 'endpoint_data', 'nodes.db']
    old_dir = os.path.dirname(os.path.realpath(__file__))
    for file in files:
        curr_path = os.path.join(old_dir, file)
        if os.path.isfile(curr_path) and os.path.isfile(curr_path):
            new_path = os.path.join(CACHE_PATH, file)
            try:
                if not os.path.exists(new_path):
                    logger.info('Moving file "%s" from "%s" to "%s".' % (file, old_dir, CACHE_PATH))
                    os.rename(curr_path, new_path)
            except OSError:
                logger.warning(
                    'Error moving cache file "%s" from "%s" to "%s".' % (file, old_dir, CACHE_PATH))


def resolve_remote_path_args(args: argparse.Namespace, attrs: list, incl_trash: bool=True):
    """In-place replaces certain attributes in Namespace by resolved node ID.
    :param attrs: list of attributes that may be given in absolute path form
    :param incl_trash: whether to resolve trashed files
    """
    for id_attr in attrs:
        if hasattr(args, id_attr):
            val = getattr(args, id_attr)
            if not val:
                continue
            if '/' in val:
                v = query.resolve_path(val, trash=incl_trash)
                if not v:
                    logger.critical('Could not resolve path "%s".' % val)
                    sys.exit(INVALID_ARG_RETVAL)
                logger.info('Resolved "%s" to "%s"' % (val, v))
                setattr(args, id_attr, v)
            elif len(val) != 22:
                logger.critical('Invalid ID format: "%s".' % val)
                sys.exit(INVALID_ARG_RETVAL)


def set_log_level(args: argparse.Namespace):
    format_ = '%(asctime)s.%(msecs).03d [%(levelname)s] [%(name)s] - %(message)s'
    time_fmt = '%y-%m-%d %H:%M:%S'

    if not args.verbose and not args.debug:
        logging.basicConfig(level=logging.WARNING, format=format_, datefmt=time_fmt)

    if args.verbose:
        logging.basicConfig(level=logging.INFO, format=format_, datefmt=time_fmt)
        if args.verbose > 1:
            logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)
            logging.getLogger('sqlalchemy.orm').setLevel(logging.INFO)
    elif args.debug:
        logging.basicConfig(level=logging.DEBUG, format=format_, datefmt=time_fmt)

        # these debug messages (prints) will not show up in log file
        import http.client

        http.client.HTTPConnection.debuglevel = 1

        if args.debug > 1:
            logging.getLogger('sqlalchemy.engine').setLevel(logging.DEBUG)
            logging.getLogger('sqlalchemy.orm').setLevel(logging.DEBUG)


def set_encoding(force_utf: bool=False):
    """Sets the default encoding to UTF-8 if none is set.
    :param force_utf: force UTF-8 output
    """
    enc = str.lower(sys.stdout.encoding)
    utf_flag = False

    if not enc or force_utf:
        import io

        sys.stdout = io.TextIOWrapper(sys.stdout.detach(), encoding='utf-8')
        sys.stderr = io.TextIOWrapper(sys.stderr.detach(), encoding='utf-8')
        utf_flag = True

        logger.info('Stdout/stderr encoding changed to UTF-8.')

    return utf_flag


def check_cache_age():
    last_sync = db.KeyValueStorage.get(CacheConsts.LAST_SYNC_KEY)
    if not last_sync:
        return
    last_sync = datetime.utcfromtimestamp(float(last_sync))
    age = (datetime.utcnow() - last_sync) / timedelta(days=1)
    if age > CacheConsts.MAX_AGE:
        logger.warning('Cache data may be outdated. Please sync.')
    else:
        logger.info('Last sync at %s.' % last_sync)


# noinspection PyProtectedMember
class Argument(object):
    """Simple argparse argument container"""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, subparser:argparse._ActionsContainer):
        subparser.add_argument(*self.args, **self.kwargs)


def main():
    max_ret = Argument('--max-retries', '-r', action='store', type=int, default=0,
                       help='set the maximum number of retries [default: 0]')

    opt_parser = argparse.ArgumentParser(
        prog=_app_name, formatter_class=argparse.RawTextHelpFormatter,
        epilog='Hints: \n'
               '  * Remote locations may be specified as path in most cases, e.g. "/folder/file", or via ID \n'
               '  * If you need to enter a node ID that contains a leading dash (minus) sign, '
               'precede it by two dashes and a space, e.g. \'-- -xfH...\'\n'
               '  * actions marked with [+] have optional arguments'
               '')
    opt_parser.add_argument('-v', '--verbose', action='count',
                            help='prints some info messages to stderr; use "-vv" to also get sqlalchemy info')
    opt_parser.add_argument('-d', '--debug', action='count',
                            help='prints info and debug to stderr; use "-dd" to also get sqlalchemy debug messages')
    opt_parser.add_argument('-c', '--color', default=format.ColorMode['never'],
                            choices=format.ColorMode.keys(),
                            help='"never" [default] turns coloring off, '
                                 '"always" turns coloring on '
                                 'and "auto" colors listings when stdout is a tty '
                                 '[uses the Linux-style LS_COLORS environment variable]')
    opt_parser.add_argument('-u', '--utf', action='store_true',
                            help='force utf output')
    opt_parser.add_argument('-nw', '--no-wait', action='store_true', help=argparse.SUPPRESS)

    subparsers = opt_parser.add_subparsers(title='action', dest='action')
    subparsers.required = True

    vers_sp = subparsers.add_parser('version', aliases=['v'], help='print version and exit\n')
    vers_sp.set_defaults(func=print_version_action)

    sync_sp = subparsers.add_parser('sync', aliases=['s'],
                                    help='[+] refresh node list cache; necessary for many actions')
    sync_sp.add_argument('--full', '-f', action='store_true', help='force a full sync')
    sync_sp.set_defaults(func=sync_action)

    old_sync_sp = subparsers.add_parser('old-sync', add_help=False)
    old_sync_sp.set_defaults(func=old_sync_action)

    clear_sp = subparsers.add_parser('clear-cache', aliases=['cc'],
                                     help='clear node cache [offline operation]\n\n')
    clear_sp.set_defaults(func=clear_action)

    tree_sp = subparsers.add_parser('tree', aliases=['t'],
                                    help='[+] print directory tree [offline operation]')
    tree_sp.add_argument('--include-trash', '-t', action='store_true')
    tree_sp.add_argument('node', nargs='?', default=None, help='root node for the tree')
    tree_sp.set_defaults(func=tree_action)

    list_c_sp = subparsers.add_parser('children', aliases=['ls', 'dir'],
                                      help='[+] list folder\'s children [offline operation]\n\n')
    list_c_sp.add_argument('--include-trash', '-t', action='store_true')
    list_c_sp.add_argument('--recursive', '-r', action='store_true')
    list_c_sp.add_argument('node')
    list_c_sp.set_defaults(func=children_action)

    find_sp = subparsers.add_parser('find', aliases=['f'],
                                    help='find nodes by name [offline operation] [case insensitive]')
    find_sp.add_argument('name')
    find_sp.set_defaults(func=find_action)

    find_hash_sp = subparsers.add_parser('find-md5', aliases=['fh'],
                                         help='find files by MD5 hash [offline operation]\n\n')
    find_hash_sp.add_argument('md5')
    find_hash_sp.set_defaults(func=find_md5_action)

    dummy_p = argparse.ArgumentParser().add_subparsers()
    re_dummy_sp = dummy_p.add_parser('', add_help=False)
    re_dummy_sp.add_argument('--max-connections', '-x', action='store', type=int, default=1,
                             help='set the maximum concurrent connections [default: 1]')
    max_ret.attach(re_dummy_sp)
    re_dummy_sp.add_argument('--exclude-ending', '-xe', action='append', dest='exclude_fe',
                             default=[],
                             help='exclude files whose endings match the given string, e.g. "bak" [case insensitive]')
    re_dummy_sp.add_argument('--exclude-regex', '-xr', action='append', dest='exclude_re',
                             default=[],
                             help='exclude files whose names match the given regular expression,'
                                  ' e.g. "^thumbs\.db$" [case insensitive]')

    upload_sp = subparsers.add_parser('upload', aliases=['ul'], parents=[re_dummy_sp],
                                      help='[+] file and directory upload to a remote destination')
    upload_sp.add_argument('--overwrite', '-o', action='store_true',
                           help='overwrite if local modification time is higher or local ctime is higher than remote '
                                'modification time and local/remote file sizes do not match.')
    upload_sp.add_argument('--force', '-f', action='store_true', help='force overwrite')
    upload_sp.add_argument('--deduplicate', '-d', action='store_true',
                           help='exclude duplicate files from upload')
    upload_sp.add_argument('path', nargs='+', help='a path to a local file or directory')
    upload_sp.add_argument('parent', help='remote parent folder')
    upload_sp.set_defaults(func=upload_action)

    overwrite_sp = subparsers.add_parser('overwrite', aliases=['ov'],
                                         help='overwrite file A [remote] with content of file B [local]')
    max_ret.attach(overwrite_sp)
    overwrite_sp.add_argument('node')
    overwrite_sp.add_argument('file')
    overwrite_sp.set_defaults(func=overwrite_action)

    download_sp = subparsers.add_parser('download', aliases=['dl'], parents=[re_dummy_sp],
                                        help='download a remote folder or file; will skip existing local files\n\n')
    download_sp.add_argument('node')
    download_sp.add_argument('path', nargs='?', default=None, help='local download path [optional]')
    download_sp.set_defaults(func=download_action)

    cr_fo_sp = subparsers.add_parser('create', aliases=['c', 'mkdir'],
                                     help='create folder using an absolute path\n\n')
    cr_fo_sp.add_argument('new_folder',
                          help='an absolute folder path, e.g. "/my/dir/"; trailing slash is optional')
    cr_fo_sp.set_defaults(func=create_action)

    trash_sp = subparsers.add_parser('list-trash', aliases=['lt'],
                                     help='[+] list trashed nodes [offline operation]')
    trash_sp.add_argument('--recursive', '-r', action='store_true')
    trash_sp.set_defaults(func=list_trash_action)

    m_trash_sp = subparsers.add_parser('trash', aliases=['rm'], help='move node to trash')
    m_trash_sp.add_argument('node')
    m_trash_sp.set_defaults(func=trash_action)

    rest_sp = subparsers.add_parser('restore', aliases=['re'], help='restore node from trash\n\n')
    rest_sp.add_argument('node', help='ID of the node')
    rest_sp.set_defaults(func=restore_action)

    move_sp = subparsers.add_parser('move', aliases=['mv'], help='move node A into folder B')
    move_sp.add_argument('child')
    move_sp.add_argument('parent')
    move_sp.set_defaults(func=move_action)

    rename_sp = subparsers.add_parser('rename', aliases=['rn'], help='rename a node\n\n')
    rename_sp.add_argument('node')
    rename_sp.add_argument('name')
    rename_sp.set_defaults(func=rename_action)

    res_sp = subparsers.add_parser('resolve', aliases=['rs'],
                                   help='resolve a path to a node ID\n\n')
    res_sp.add_argument('path')
    res_sp.set_defaults(func=resolve_action)

    # maybe the child operations should not be exposed
    # they can be used for creating hardlinks
    add_c_sp = subparsers.add_parser('add-child', aliases=['ac'],
                                     help='add a node to a parent folder')
    add_c_sp.add_argument('parent')
    add_c_sp.add_argument('child')
    add_c_sp.set_defaults(func=add_child_action)

    rem_c_sp = subparsers.add_parser('remove-child', aliases=['rc'],
                                     help='remove a node from a parent folder\n\n')
    rem_c_sp.add_argument('parent')
    rem_c_sp.add_argument('child')
    rem_c_sp.set_defaults(func=remove_child_action)

    usage_sp = subparsers.add_parser('usage', aliases=['u'], help='show drive usage data')
    usage_sp.set_defaults(func=usage_action)

    quota_sp = subparsers.add_parser('quota', aliases=['q'], help='show drive quota [raw JSON]')
    quota_sp.set_defaults(func=quota_action)

    meta_sp = subparsers.add_parser('metadata', aliases=['m'],
                                    help='print a node\'s metadata [raw JSON]\n\n')
    meta_sp.add_argument('node')
    meta_sp.set_defaults(func=metadata_action)

    de_sp = subparsers.add_parser('delete-everything', add_help=False)
    de_sp.set_defaults(func=delete_everything_action)

    # useful for interactive mode
    dn_sp = subparsers.add_parser('init', aliases=['i'], add_help=False)
    dn_sp.set_defaults(func=None)

    # dump sql database creation sequence to stdout
    dmp_sp = subparsers.add_parser('dumpsql', add_help=False)
    dmp_sp.set_defaults(func=dump_sql_action)

    fuse_sp = subparsers.add_parser('mount', help='[+] mount the cloud drive at a local directory')
    fuse_sp.add_argument('--ro', '-ro', action='store_true', help='mount read-only')
    fuse_sp.add_argument('--foreground', '-fg', action='store_true', help='do not detach')
    fuse_sp.add_argument('path')
    fuse_sp.set_defaults(func=mount_action)

    plugin_log = [str(plugins.Plugin)]
    for plugin in plugins.Plugin:
        if plugin.check_version(__version__):
            log = []
            plugin.attach(subparsers, log)
            plugin_log.extend(log)
        else:
            plugin_log.append('Script version is not compatible with "%s".' % plugin)

    args = opt_parser.parse_args()

    set_log_level(args)
    for msg in plugin_log:
        logger.info(msg)

    set_encoding(force_utf=args.utf)

    migrate_cache_files()

    if args.func not in offline_actions:
        if not common.init(CACHE_PATH):
            sys.exit(INIT_FAILED_RETVAL)

    if args.func not in nocache_actions:
        if not db.init(CACHE_PATH):
            sys.exit(INIT_FAILED_RETVAL)
        check_cache_age()

    format.init(args.color)

    if args.no_wait:
        common.BackOffRequest._wait = lambda: None

    autoresolve_attrs = ['child', 'parent', 'node']
    resolve_remote_path_args(args, autoresolve_attrs,
                             incl_trash=args.action not in no_autores_trash_actions)

    # call appropriate sub-parser action
    if args.func:
        sys.exit(args.func(args))


if __name__ == "__main__":
    main()



__version__ = '0.7.1'
__all__ = ('account', 'common', 'content', 'metadata', 'trash')

"""
================
Node JSON Format
================

This is the usual node JSON format for a file::
    {
           'contentProperties': {'contentType': 'text/plain',
                                 'extension': 'txt',
                                 'md5': 'd41d8cd98f00b204e9800998ecf8427e',
                                 'size': 0,
                                 'version': 1},
           'createdBy': '<security-profile-nm>-<user>',
           'createdDate': '2015-01-01T00:00:00.00Z',
           'eTagResponse': 'AbCdEfGhI01',
           'id': 'AbCdEfGhIjKlMnOpQr0123',
           'isShared': False,
           'kind': 'FILE',
           'labels': [],
           'modifiedDate': '2015-01-01T00:00:00.000Z',
           'name': 'empty.txt',
           'parents': ['0123AbCdEfGhIjKlMnOpQr'],
           'restricted': False,
           'status': 'AVAILABLE',
           'version': 1
    }

The ``modifiedDate`` and ``version`` keys get updated each time the content or metadata is updated.
``contentProperties['version']`` gets updated on overwrite.

A folder's JSON looks similar, but it lacks the ``contentProperties`` dictionary.

"""


import hashlib
import logging
import os

logger = logging.getLogger(__name__)


class IncrementalHasher(object):
    __slots__ = ('hasher',)

    def __init__(self):
        self.hasher = hashlib.md5()

    def update(self, chunk):
        self.hasher.update(chunk)

    def get_result(self) -> str:
        return self.hasher.hexdigest()

def hash_bytes(bytes_) -> str:
    hasher = hashlib.md5()
    bytes_.seek(0)
    for chunk in iter(lambda: bytes_.read(1024 ** 2), b''):
        hasher.update(chunk)
    return hasher.hexdigest()

def hash_file(file_name: str) -> str:
    with open(file_name, 'rb') as f:
        md5 = hash_bytes(f)
    logger.debug('MD5 of "%s" is %s' % (os.path.basename(file_name), md5))
    return md5


# encoder.py
# taken from requests_toolbelt
# Apache license
# ff96b05243a3f5d4689cb776f4d03c9bda041957
#

# fuse.py
# ISC license
# https://github.com/akovari/fusepy/commit/847388cd52f060d5678980f20043afeedd166a55


import logging
import collections

from .common import *

logger = logging.getLogger(__name__)


class _Usage(object):
    dict_ = {}

    def __init__(self, dict_):
        self.dict_ = dict_

    @staticmethod
    def format_line(type_, count, size):
        return '{0:10} {1:7}, {2:>6} {3:3}\n'.format(type_ + ':', count, *size)

    def __str__(self):
        str_ = ''
        try:
            sum_count = 0
            sum_bytes = 0
            for key in self.dict_.keys():
                if not isinstance(self.dict_[key], dict):
                    continue
                sum_count += self.dict_[key]['total']['count']
                sum_bytes += self.dict_[key]['total']['bytes']
            types = collections.OrderedDict([('Documents', 'doc'),
                                             ('Other', 'other'),
                                             ('Photos', 'photo'),
                                             ('Videos', 'video')])
            total_count = 0
            total_bytes = 0
            for desc in types:
                t = types[desc]
                type_usage = self.dict_[t]['total']
                type_count = type_usage['count']
                type_bytes = type_usage['bytes']
                total_count += type_count
                total_bytes += type_bytes
                str_ += _Usage.format_line(desc, type_count, _Usage.file_size_pair(type_bytes))
            str_ += _Usage.format_line('Total', total_count, _Usage.file_size_pair(total_bytes))
        except KeyError:
            logger.warning('Invalid usage JSON string.')
        return str_

    @staticmethod
    def file_size_pair(num: int, suffix='B') -> str:
        for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
            if abs(num) < 1024.0:
                return '%3.1f' % num, '%s%s' % (unit, suffix)
            num /= 1024.0
        return '%.1f' % num, '%s%s' % ('Yi', suffix)


def get_account_info() -> dict:
    r = BackOffRequest.get(get_metadata_url() + 'account/info')
    return r.json()


def get_account_usage() -> str:
    r = BackOffRequest.get(get_metadata_url() + 'account/usage')
    if r.status_code not in OK_CODES:
        raise RequestError(r.status_code, r.text)
    return _Usage(r.json())


def get_quota() -> dict:
    r = BackOffRequest.get(get_metadata_url() + 'account/quota')
    if r.status_code not in OK_CODES:
        raise RequestError(r.status_code, r.text)
    return r.json()


def fs_sizes() -> tuple:
    """returns total and free space"""
    q = get_quota()
    return q.get('quota', 0), q.get('available', 0)



"""
Syncs Amazon Node API objects with sqlite database
"""

import logging
from sqlalchemy.exc import *
from datetime import datetime, timedelta

try:
    import dateutil.parser as iso_date
except ImportError:
    # noinspection PyPep8Naming
    class iso_date(object):
        @staticmethod
        def parse(str_: str):
            return datetime.strptime(str_, '%Y-%m-%dT%H:%M:%S.%fZ')

from . import db

logger = logging.getLogger(__name__)


def remove_purged(purged: list):
    session = db.Session()
    for p_id in purged:
        session.query(db.Node).filter_by(id=p_id).delete()
    session.commit()
    logger.info('Purged %i nodes.' % len(purged))


def insert_nodes(nodes: list, partial=True):
    """Inserts mixed list of files and folders into cache."""
    files = []
    folders = []
    for node in nodes:
        if node['status'] == 'PENDING':
            continue
        kind = node['kind']
        if kind == 'FILE':
            files.append(node)
        elif kind == 'FOLDER':
            folders.append(node)
        elif kind != 'ASSET':
            logger.warning('Cannot insert unknown node type "%s".' % kind)
    insert_folders(folders)
    insert_files(files)

    insert_parentage(files + folders, partial)

def insert_node(node: db.Node):
    """Inserts single file or folder into cache."""
    if not node:
        pass
    insert_nodes([node])


def insert_folders(folders: list):
    """ Inserts list of folders into cache. Sets 'update' column to current date.
    :param folders: list of raw dict-type folders
    """

    if not folders:
        return

    session = db.Session()
    for folder in folders:
        logger.debug(folder)

        # root folder has no name key
        f_name = folder.get('name')
        f = db.Folder(folder['id'], f_name,
                      iso_date.parse(folder['createdDate']),
                      iso_date.parse(folder['modifiedDate']),
                      folder['status'])
        f.updated = datetime.utcnow()
        session.merge(f)

    try:
        session.commit()
    except IntegrityError:
        logger.warning('Error inserting folders.')
        session.rollback()

    logger.info('Inserted/updated %d folders.' % len(folders))

def insert_files(files: list):
    if not files:
        return

    stmt1 = str(db.Node.__table__.insert())
    stmt1 = stmt1.replace('INSERT INTO', 'INSERT OR REPLACE INTO')

    stmt2 = str(db.File.__table__.insert())
    stmt2 = stmt2.replace('INSERT INTO', 'INSERT OR REPLACE INTO')

    db.engine.execute(
        stmt1,
        [dict(id=f['id'], type='file', name=f.get('name'), description=f.get('description'),
              created=iso_date.parse(f['createdDate']),
              modified=iso_date.parse(f['modifiedDate']),
              updated=datetime.utcnow(),
              status=f['status']
              ) for f in files
         ]
    )
    db.engine.execute(
        stmt2,
        [dict(id=f['id'], md5=f.get('contentProperties', {}).get('md5', 'd41d8cd98f00b204e9800998ecf8427e'),
              size=f.get('contentProperties', {}).get('size', 0)
              ) for f in files
         ]
    )

    logger.info('Inserted/updated %d files.' % len(files))

def insert_parentage(nodes: list, partial=True):
    conn = db.engine.connect()
    trans = conn.begin()

    if partial:
        conn.execute('DELETE FROM parentage WHERE child IN (%s)' %
                     ', '.join(['"' + n['id'] + '"' for n in nodes]))
    for n in nodes:
        for p in n['parents']:
            conn.execute('INSERT OR IGNORE INTO parentage VALUES (?, ?)', p, n['id'])
    trans.commit()

    logger.info('Parented %d nodes.' % len(nodes))


import argparse


class RegisterLeafClasses(type):
    def __init__(cls, name, bases, nmspc):
        super(RegisterLeafClasses, cls).__init__(name, bases, nmspc)
        if not hasattr(cls, 'registry'):
            cls.registry = set()
        cls.registry.add(cls)
        cls.registry -= set(bases) # Remove base classes

    # metamethods, called on class objects:
    def __iter__(cls):
        return iter(cls.registry)

    def __str__(cls):
        if cls in cls.registry:
            return cls.__name__
        return cls.__name__ + " leaf classes: " + ", ".join([sc.__name__ for sc in cls])


class Plugin(object, metaclass=RegisterLeafClasses):
    """Plugin base class. May be subject to changes."""
    MIN_VERSION = None
    MAX_VERSION = None

    @classmethod
    def check_version(cls, version: str) -> bool:
        from distutils.version import StrictVersion
        if cls.MIN_VERSION:
            if StrictVersion(cls.MIN_VERSION) > StrictVersion(version):
                return False
        if cls.MAX_VERSION:
            return StrictVersion(cls.MAX_VERSION) >= StrictVersion(version)
        return True

    @classmethod
    def __str__(cls):
        return cls.__name__

    @classmethod
    def attach(cls, subparsers: argparse.ArgumentParser, log: list, **kwargs):
        pass

    @staticmethod
    def action(args: argparse.Namespace):
        pass


import os
import re
from setuptools import setup, find_packages
from distutils.version import StrictVersion


def read(fname: str) -> str:
    return open(os.path.join(os.path.dirname(__file__), fname), encoding='utf-8').read()


version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                    read('acd_cli.py'), re.MULTILINE).group(1)
StrictVersion(version)

setup(
    name='acdcli',
    version='0.3.0a1',
    description='a command line interface for Amazon Cloud Drive',
    long_description=read('README.rst').replace('✅', '✓'),
    license='GPLv2+',
    author='yadayada',
    author_email='acd_cli@mail.com',
    keywords='amazon cloud drive clouddrive',
    url='https://github.com/yadayada/acd_cli',
    download_url='https://github.com/yadayada/acd_cli/tarball/0.3.0a1',
    zip_safe=False,
    packages=find_packages(exclude=['tests']),
    test_suite='tests.get_suite',
    scripts=['acd_cli.py'],
    entry_points={'console_scripts': ['acd_cli = acd_cli:main', 'acdcli = acd_cli:main'],
                  # 'acd_cli.plugins': ['stream = plugins.stream',
                  # 'template = plugins.template']
                  },
    install_requires=['appdirs', 'python_dateutil', 'requests>=2.1.0', 'requests_toolbelt',
                      'sqlalchemy'],
    tests_require=['httpretty', 'mock'],
    classifiers=[
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',
        'Development Status :: 4 - Beta',
        'Topic :: System :: Archiving :: Backup',
        'Topic :: System :: Filesystems'
    ]
)



"""Experimental fusepy read support"""

import os
import stat
import errno
import logging
from time import time
from datetime import datetime, timedelta
from collections import deque, defaultdict

from acdcli.bundled.fuse import FUSE, FuseOSError, Operations
from acdcli.cache import query, sync
from acdcli.api import account, content, metadata, trash
from acdcli.api.common import RequestError

logger = logging.getLogger(__name__)

FUSE_BS = 128 * 1024
CHUNK_SZ = content.CHUNK_SIZE


class StreamedResponseCache(object):
    """Stream response cache intended for consecutive access."""

    class StreamChunk(object):
        __slots__ = ('offset', 'r', 'end')

        def __init__(self, id, offset, length, **kwargs):
            self.offset = offset
            self.r = content.response_chunk(id, offset, length, **kwargs)
            self.end = offset + int(self.r.headers['content-length']) - 1

        def has_byte_range(self, offset, length):
            logger.debug('s: %d-%d; r: %d-%d'
                         % (self.offset, self.end, offset, offset + length - 1))
            if offset == self.offset and offset + length - 1 <= self.end:
                return True

        def get(self, length):
            b = next(self.r.iter_content(length))
            self.offset += len(b)
            logger.debug(len(b))
            return b

        def close(self):
            self.r.close()

    class File(object):
        __slots__ = ('chunks', 'access')

        def __init__(self):
            self.chunks = deque(maxlen=15)
            self.access = time()

        def get(self, id, offset, length, total):
            self.access = time()

            for c in self.chunks:
                if c.has_byte_range(offset, length):
                    try:
                        bytes_ = c.get(length)
                    except:
                        self.chunks.remove(c)
                    else:
                        return bytes_
            try:
                chunk = StreamedResponseCache. \
                    StreamChunk(id, offset, CHUNK_SZ, timeout=5)
            except RequestError as e:
                raise FuseOSError(errno.ECOMM)
            else:
                self.chunks.append(chunk)
                return chunk.get(length)

        def clear(self):
            for chunk in self.chunks():
                try:
                    chunk.close()
                except:
                    pass

    files = defaultdict(File)

    @classmethod
    def get(cls, id, offset, length, total):
        return cls.files[id].get(id, offset, length, total)

    @classmethod
    def invalidate(cls):
        pass


class ACDFuse(Operations):
    def __init__(self):
        self.total, _ = account.fs_sizes()
        self.free = self.total - query.calculate_usage()
        self.fh = 1

    def readdir(self, path, fh):
        logger.debug('readdir %s' % path)

        node, _ = query.resolve(path, trash=False)
        if not node:
            raise FuseOSError(errno.ENOENT)

        return [_ for _ in ['.', '..'] + [c.name for c in node.available_children()]]

    def getattr(self, path, fh=None):
        logger.debug('getattr %s' % path)

        id = query.resolve_path(path, trash=False)
        node = query.get_node(id)
        if not node:
            raise FuseOSError(errno.ENOENT)

        times = dict(st_atime=time(),
                     st_mtime=(node.modified - datetime(1970, 1, 1)) / timedelta(seconds=1),
                     st_ctime=(node.created - datetime(1970, 1, 1)) / timedelta(seconds=1))

        if node.is_folder():
            nlinks = 1 + len(node.parents)
            for c in node.children:
                if c.is_folder() and c.is_available():
                    nlinks += 1
            return dict(st_mode=stat.S_IFDIR | 0o0777, st_nlink=nlinks, **times)
        if node.is_file():
            return dict(st_mode=stat.S_IFREG | 0o0666,
                        st_nlink=len(node.parents), st_size=node.size, **times)

    def read(self, path, length, offset, fh):
        logger.debug("read %s, ln: %d of: %d fh %d" % (os.path.basename(path), length, offset, fh))
        node, _ = query.resolve(path, trash=False)
        if node.size == 0:
            return b''

        return StreamedResponseCache.get(node.id, offset, length, node.size)

    def statfs(self, path):
        bs = 512 * 1024
        return dict(f_bsize=bs,
                    f_frsize=bs,
                    f_blocks=self.total // bs,  # total no of blocks
                    f_bfree=self.free // bs,  # free blocks
                    f_bavail=self.free // bs,
                    f_namemax=256
                    )

    def mkdir(self, path, mode):
        logger.debug('+mkdir %s' % path)

        name = os.path.basename(path)
        ppath = os.path.dirname(path)
        pid = query.resolve_path(ppath)
        if not pid:
            raise FuseOSError(errno.ENOTDIR)

        try:
            r = content.create_folder(name, pid)
        except RequestError as e:
            if e.status_code == e.CODE.CONN_EXCEPTION:
                raise FuseOSError(errno.ECOMM)
            elif e.status_code == 409:
                raise FuseOSError(errno.EEXIST)
            else:
                raise FuseOSError(errno.EREMOTEIO)
        else:
            sync.insert_node(r)

    @staticmethod
    def _trash(path):
        logger.debug('trash %s' % path)
        node, parent = query.resolve(path)

        if not node:  # or not parent:
            raise FuseOSError(errno.ENOENT)

        logger.debug('%s %s' % (node, parent))

        try:
            # if len(node.parents) > 1:
            #     r = metadata.remove_child(parent.id, node.id)
            # else:
            r = trash.move_to_trash(node.id)
        except RequestError as e:
            if e.status_code == e.CODE.CONN_EXCEPTION:
                raise FuseOSError(errno.ECOMM)
            else:
                raise FuseOSError(errno.EREMOTEIO)
        else:
            sync.insert_node(r)

    def rmdir(self, path):
        logger.debug('-rmdir %s' % path)
        self._trash(path)

    def unlink(self, path):
        logger.debug('-unlink %s' % path)
        self._trash(path)

    def create(self, path, mode):
        logger.debug('+create %s' % path)

        name = os.path.basename(path)
        ppath = os.path.dirname(path)
        pid = query.resolve_path(ppath)
        if not pid:
            raise FuseOSError(errno.ENOTDIR)

        try:
            r = content.create_file(name, pid)
            sync.insert_node(r)
        except RequestError as e:
            if e.status_code == e.CODE.CONN_EXCEPTION:
                raise FuseOSError(errno.ECOMM)
            elif e.status_code == 409:
                raise FuseOSError(errno.EEXIST)
            else:
                raise FuseOSError(errno.EREMOTEIO)

        self.fh += 1
        return self.fh

    def rename(self, old, new):
        if old == new:
            return

        logger.debug('rename %s %s' % (old, new))

        id = query.resolve_path(old)
        if not id:
            raise FuseOSError(errno.ENOENT)

        new_bn, old_bn = os.path.basename(new), os.path.basename(old)
        new_dn, old_dn = os.path.dirname(new), os.path.dirname(old)

        existing_id = query.resolve_path(new)
        if existing_id:
            en = query.get_node(existing_id)
            if en and en.is_file() and en.size == 0:
                trash.move_to_trash(existing_id)
            else:
                raise FuseOSError(errno.EEXIST)

        if new_bn != old_bn:
            self._rename(id, new_bn)

        if new_dn != old_dn:
            odir_id = query.resolve_path(old_dn)
            ndir_id = query.resolve_path(new_dn)
            if not odir_id or not ndir_id:
                raise FuseOSError(errno.ENOTDIR)
            self._move(id, odir_id, ndir_id)

    @staticmethod
    def _rename(id, name):
        try:
            r = metadata.rename_node(id, name)
        except RequestError as e:
            logger.debug(e)
            if e.status_code == e.CODE.CONN_EXCEPTION:
                raise FuseOSError(errno.ECOMM)
            elif e.status_code == 409:
                raise FuseOSError(errno.EEXIST)
            else:
                raise FuseOSError(errno.EREMOTEIO)
        else:
            sync.insert_node(r)

    @staticmethod
    def _move(id, old_folder, new_folder):
        try:
            r = metadata.move_node_from(id, old_folder, new_folder)
        except RequestError as e:
            if e.status_code == e.CODE.CONN_EXCEPTION:
                raise FuseOSError(errno.ECOMM)
            raise FuseOSError(errno.EREMOTEIO)
        else:
            sync.insert_node(r)

    def open(self, path, flags):
        logger.debug('open %s' % path)
        self.fh += 1
        return self.fh

    # def write(self, path, data, offset, fh):
    #     logger.debug('write %s , l: %d, o: %d, fh: %d' % (path, 0, offset, fh))
    #     # with open(os.devnull, 'wb') as dn:
    #     return len(data)

    # def truncate(self, path, length, fh=None):
    #     logger.debug('truncate %s path, %d' % (path, length))
    #     pass

    def release(self, path, fh):
        logger.debug('release %s, %d' % (path, fh))

    def utimens(self, path, times=None):
        logger.debug('utimens %s' % path)

    def chmod(self, path, mode):
        logger.debug('chmod %s %s' % (path, oct(mode)))

    def chown(self, path, uid, gid):
        logger.debug('chmod %s %d %d' % (path, uid, gid))


def mount(path: str, **kwargs):
    if not query.get_root_node():
        logger.critical('Root node not found. Aborting.')
        return
    if not os.path.isdir(path):
        logger.critical('Mount directory does not exist.')
        return

    FUSE(ACDFuse(), path, entry_timeout=60, auto_cache=True,
         nothreads=True,  # threading will break the caching
         uid=os.getuid(), gid=os.getgid(),
         **kwargs
         )
